# https://nomads.ncep.noaa.gov/gribfilter.php?ds=gfs_0p25_1hr
gfs_parameters_dict = {
    'var_4LFTX': 'off',
    'var_ABSV': 'off',
    'var_ACPCP': 'off',
    'var_ALBDO': 'off',
    'var_APCP': 'off',
    'var_CAPE': 'off',
    'var_CFRZR': 'off',
    'var_CICEP': 'off',
    'var_CIN': 'off',
    'var_CLWMR': 'off',
    'var_CNWAT': 'off',
    'var_CPOFP': 'off',
    'var_CPRAT': 'off',
    'var_CRAIN': 'off',
    'var_CSNOW': 'off',
    'var_CWAT': 'off',
    'var_CWORK': 'off',
    'var_DLWRF': 'off',
    'var_DPT': 'off',
    'var_DSWRF': 'off',
    'var_DZDT': 'off',
    'var_FLDCP': 'off',
    'var_FRICV': 'off',
    'var_GFLUX': 'off',
    'var_GRLEV': 'off',
    'var_GUST': 'off',
    'var_HCDC': 'off',
    'var_HGT': 'off',
    'var_HINDEX': 'off',
    'var_HLCY': 'off',
    'var_HPBL': 'off',
    'var_ICAHT': 'off',
    'var_ICEC': 'off',
    'var_ICEG': 'off',
    'var_ICETK': 'off',
    'var_ICETMP': 'off',
    'var_ICMR': 'off',
    'var_LAND': 'off',
    'var_LCDC': 'off',
    'var_LFTX': 'off',
    'var_LHTFL': 'off',
    'var_MCDC': 'off',
    'var_MSLET': 'off',
    'var_O3MR': 'off',
    'var_PEVFR': 'off',
    'var_PLPL': 'off',
    'var_POT': 'off',
    'var_PRATE': 'off',
    'var_PRES': 'off',
    'var_PRMSL': 'off',
    'var_PWAT': 'off',
    'var_REFC': 'off',
    'var_REFD': 'off',
    'var_RH': 'off',
    'var_RWMR': 'off',
    'var_SFCR': 'off',
    'var_SHTFL': 'off',
    'var_SNMR': 'off',
    'var_SNOD': 'off',
    'var_SOILL': 'off',
    'var_SOILW': 'off',
    'var_SOTYP': 'off',
    'var_SPFH': 'off',
    'var_SUNSD': 'off',
    'var_TCDC': 'off',
    'var_TMAX': 'off',
    'var_TMIN': 'off',
    'var_TMP': 'off',
    'var_TOZNE': 'off',
    'var_TSOIL': 'off',
    'var_UFLX': 'off',
    'var_UGRD': 'off',
    'var_U-GWD': 'off',
    'var_ULWRF': 'off',
    'var_USTM': 'off',
    'var_USWRF': 'off',
    'var_VEG': 'off',
    'var_VFLX': 'off',
    'var_VGRD': 'off',
    'var_V-GWD': 'off',
    'var_VIS': 'off',
    'var_VRATE': 'off',
    'var_VSTM': 'off',
    'var_VVEL': 'off',
    'var_VWSH': 'off',
    'var_WATR': 'off',
    'var_WEASD': 'off',
    'var_WILT': 'off',
}

gfs_levels_dict = {
    'lev_0-0.1_m_below_ground': 'off',
    'lev_0.1-0.4_m_below_ground': 'off',
    'lev_0.4-1_m_below_ground': 'off',
    'lev_1-2_m_below_ground': 'off',
    'lev_2_m_above_ground': 'off',
    'lev_10_m_above_ground': 'off',
    'lev_20_m_above_ground': 'off',
    'lev_30_m_above_ground': 'off',
    'lev_40_m_above_ground': 'off',
    'lev_50_m_above_ground': 'off',
    'lev_80_m_above_ground': 'off',
    'lev_100_m_above_ground': 'off',
    'lev_1000_m_above_ground': 'off',
    'lev_4000_m_above_ground': 'off',
    'lev_10_m_above_mean_sea_level': 'off',
    'lev_1829_m_above_mean_sea_level': 'off',
    'lev_2743_m_above_mean_sea_level': 'off',
    'lev_3658_m_above_mean_sea_level': 'off',
    'lev_3000-0_m_above_ground': 'off',
    'lev_6000-0_m_above_ground': 'off',
    'lev_180-0_mb_above_ground': 'off',
    'lev_255-0_mb_above_ground': 'off',
    'lev_90-0_mb_above_ground': 'off',
    'lev_30-0_mb_above_ground': 'off',
    'lev_0C_isotherm': 'off',
    'lev_1_hybrid_level': 'off',
    'lev_2_hybrid_level': 'off',
    'lev_1000_mb': 'off',
    'lev_975_mb': 'off',
    'lev_950_mb': 'off',
    'lev_925_mb': 'off',
    'lev_900_mb': 'off',
    'lev_850_mb': 'off',
    'lev_800_mb': 'off',
    'lev_750_mb': 'off',
    'lev_700_mb': 'off',
    'lev_650_mb': 'off',
    'lev_600_mb': 'off',
    'lev_550_mb': 'off',
    'lev_500_mb': 'off',
    'lev_450_mb': 'off',
    'lev_400_mb': 'off',
    'lev_350_mb': 'off',
    'lev_300_mb': 'off',
    'lev_250_mb': 'off',
    'lev_200_mb': 'off',
    'lev_150_mb': 'off',
    'lev_100_mb': 'off',
    'lev_70_mb': 'off',
    'lev_50_mb': 'off',
    'lev_40_mb': 'off',
    'lev_30_mb': 'off',
    'lev_20_mb': 'off',
    'lev_15_mb': 'off',
    'lev_10_mb': 'off',
    'lev_7_mb': 'off',
    'lev_5_mb': 'off',
    'lev_3_mb': 'off',
    'lev_2_mb': 'off',
    'lev_1_mb': 'off',
    'lev_0.7_mb': 'off',
    'lev_0.4_mb': 'off',
    'lev_0.2_mb': 'off',
    'lev_0.1_mb': 'off',
    'lev_0.07_mb': 'off',
    'lev_0.04_mb': 'off',
    'lev_0.02_mb': 'off',
    'lev_0.01_mb': 'off',
    'lev_surface': 'off',
    'lev_mean_sea_level': 'off',
    'lev_boundary_layer_cloud_layer': 'off',
    'lev_planetary_boundary_layer': 'off',
    'lev_tropopause': 'off',
    'lev_entire_atmosphere': 'off',
    'lev_entire_atmosphere_(considered_as_a_single_layer)': 'off',
    'lev_top_atmosphere': 'off',
    'lev_convective_cloud_layer': 'off',
    'lev_convective_bottom_level': 'off',
    'lev_convective_top_level': 'off',
    'lev_high_cloud_layer': 'off',
    'lev_high_cloud_bottom_level': 'off',
    'lev_high_cloud_top_level': 'off',
    'lev_low_cloud_layer': 'off',
    'lev_low_cloud_bottom_level': 'off',
    'lev_low_cloud_top_level': 'off',
    'lev_middle_cloud_layer': 'off',
    'lev_middle_cloud_bottom_level': 'off',
    'lev_middle_cloud_top_level': 'off',
    'lev_cloud_ceiling': 'off',
    'lev_highest_tropospheric_freezing_level': 'off',
    'lev_PV%3D-2e-06_(Km%5E2%2Fkg%2Fs)_surface': 'off',
    'lev_PV%3D2e-06_(Km%5E2%2Fkg%2Fs)_surface': 'off',
}



"""
Author: Wenyu Ouyang
Date: 2023-10-31 20:56:23
LastEditTime: 2024-03-28 08:39:15
LastEditors: Wenyu Ouyang
Description: Test funcs for reading stations' data
FilePath: \hydrodata\tests\test_reader_stations.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pandas as pd

from hydrodatasource.reader.stations import huanren_preprocess


def test_huanren_preprocess(tmp_path):
    # create a temporary directory for testing
    test_dir = tmp_path / "test_data"
    test_dir.mkdir()

    # create a test excel file
    test_file = test_dir / "test_file.xlsx"
    test_data = pd.DataFrame(
        {"date": ["2022-01-01", "2022-01-02", "2022-01-03"], "flow": [1.0, 2.0, 3.0]}
    )
    test_data.to_excel(test_file, index=False)

    # set up local data path and run function
    os.environ["LOCAL_DATA_PATH"] = str(test_dir)
    huanren_preprocess()

    # check that the output file was created and has the correct data
    output_file = test_dir / "huanren" / "tidy.csv"
    assert output_file.exists()
    output_data = pd.read_csv(output_file)
    assert output_data.equals(test_data)



from hydrodatasource.configs.config import PS
from datetime import datetime

def read_data(stcd, datatype, start_time, end_time):
    ps = PS.cursor()
    dt_start = datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S")
    dt_end = datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%S")
    dt_start_time = dt_start.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
    dt_end_time = dt_end.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
    if datatype == 'rain':
        table = 'st_pptn_r'
        columns = "tm, intv, drp"
    elif datatype == 'streamflow':
        table = 'st_rsvr_r'
        columns = "tm, inq"
    else:
        raise ValueError("datatype must be 'rain' or 'streamflow'")
    sql = f"SELECT {columns} FROM {table} where stcd = %s and tm between %s and %s"
    PS.autocommit = True
    ps.execute(sql, (stcd, dt_start_time, dt_end_time))
    return ps.fetchall()


import collections
import os
from abc import ABC
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import xarray as xr
from hydrodataset import HydroDataset
from hydroutils import hydro_file, hydro_time
from tqdm import tqdm

import hydrodatasource.configs.config as conf
from hydrodatasource.configs.config import SETTING
from hydrodatasource.reader import access_fs
from hydrodatasource.reader.reader import DataHandler

CACHE_DIR = hydro_file.get_cache_dir()


class HydroData(ABC):
    """An interface for reading multi-modal data sources.

    Parameters
    ----------
    ABC : _type_
        _description_
    """

    def __init__(self, data_path):
        self.data_source_dir = data_path

    def get_name(self):
        raise NotImplementedError

    def set_data_source_describe(self):
        raise NotImplementedError

    def read_data(self):
        raise NotImplementedError


class SelfMadeCamels(HydroDataset):
    """A class for reading hydrodataset, but not really ready-datasets,
    just some data directorys organized like a CAMELS dataset.

    Typically, we read our self-made data.

    NOTE:
    open data: Mainly for baisn from GAGES-II (FOR US) or GRDC (FOR CHINA)
    private data: for stations with streamflow data

    No matter open or private, we will compile forcing data and attr data into a dataset,
    organized like a ready dataset -- like CAMELS, as these are for paper publication.

    The process of compiling forcing data and attr data into these streamflow dataset is
    not here. It is in some preprocessing scripts such as CatchmentAttributes repo: https://github.com/OuyangWenyu/CatchmentAttributes
    or CatchmentForcing repo: https://github.com/OuyangWenyu/CatchmentForcings.

    They are generally not full automatic, as some data are not open, and some data are not ready for use.

    For streamflow data, we will transform them to netcdf format, and combine them with forcing data into a timesereis dataset.

    Hence, in this module, we directly read the compiled dataset: one time-series nc file and one attr nc file.
    If any data is not ready, we will raise an error and preprocess in other scripts again.
    """

    def __init__(self, data_path, download=False, version="latest"):
        """Initialize a self-made CAMELS dataset.

        Parameters
        ----------
        data_path : _type_
            _description_
        download : bool, optional
            _description_, by default False
        version : str, optional
            We have multiple versions of self-made CAMELS dataset, by default "latest"
        """
        super().__init__(data_path)
        # the naming convention for basin ids are needed
        # we use GRDC station's ids as our default coding convention
        # GRDC station ids are 7 digits, the first 1 digit is continent code,
        # the second 4 digits are sub-region related code
        self.data_source_description = self.set_data_source_describe()
        if download:
            self.download_data_source()
        self.camels_sites = self.read_site_info()
        # for camels_cc (version 1), {"time": "DATE", "streamflow": "Q"}
        # Here the dict is for camels_cc_v2
        if version in ["latest", "v2"]:
            self.VAR_DICT = {"time": "time", "streamflow": "streamflow"}
        elif version == "v1":
            self.VAR_DICT = {"time": "DATE", "streamflow": "Q"}
        else:
            raise ValueError("version must be latest, v2, or v1")

    @property
    def streamflow_unit(self):
        return "m^3/s"

    def get_name(self):
        return "SelfMadeCamels"

    def set_data_source_describe(self):
        camels_db = self.data_source_dir
        # shp files of basins
        camels_shp_files_dir = os.path.join(camels_db, "basin_boudaries")
        # attr, flow and forcing data are all in the same dir. each basin has one dir.
        flow_dir = os.path.join(camels_db, "streamflow")
        sm_dir = os.path.join(camels_db, "soil_moisture")
        et_dir = os.path.join(camels_db, "evapotranspiration")
        forcing_dir = os.path.join(camels_db, "basin_mean_forcing")
        attr_dir = os.path.join(camels_db, "attribute")
        # no gauge id file for CAMELS_CC, just read from any attribute file
        gauge_id_file = os.path.join(camels_db, "gage_points.csv")
        attr_key_lst = [
            "climate",
            "geology",
            "land_cover",
            "permeability_porosity",
            "root_depth",
            "soil",
            "topo_elev_slope",
            "topo_shape_factors",
        ]
        return collections.OrderedDict(
            CAMELS_DIR=camels_db,
            CAMELS_FLOW_DIR=flow_dir,
            CAMELS_SM_DIR=sm_dir,
            CAMELS_ET_DIR=et_dir,
            CAMELS_FORCING_DIR=forcing_dir,
            CAMELS_ATTR_DIR=attr_dir,
            CAMELS_ATTR_KEY_LST=attr_key_lst,
            CAMELS_GAUGE_FILE=gauge_id_file,
            CAMELS_BASINS_SHP_DIR=camels_shp_files_dir,
        )

    def download_data_source(self):
        print(
            "Please download it manually and put all files of a CAMELS dataset in the CAMELS_DIR directory."
        )
        print("We unzip all files now.")

    def read_site_info(self):
        camels_file = self.data_source_description["CAMELS_GAUGE_FILE"]
        return pd.read_csv(camels_file, sep=",", dtype={"gage_id": str})

    def read_object_ids(self, object_params=None) -> np.array:
        return self.camels_sites["gage_id"].values

    def read_target_cols(
        self, object_ids=None, t_range_list=None, target_cols=None, **kwargs
    ) -> np.array:
        if target_cols is None:
            return np.array([])
        else:
            nf = len(target_cols)
        t_range_list = hydro_time.t_range_days(t_range_list)
        nt = t_range_list.shape[0]
        y = np.full([len(object_ids), nt, nf], np.nan)
        streamflow_name = self.VAR_DICT["streamflow"]
        time_name = self.VAR_DICT["time"]
        for j in tqdm(
            range(len(target_cols)), desc="Read streamflow data of CAMELS-CC"
        ):
            for k in tqdm(range(len(object_ids))):
                # only one streamflow type: Q
                flow_file = os.path.join(
                    self.data_source_description["CAMELS_FLOW_DIR"],
                    object_ids[k] + ".csv",
                )
                flow_data = pd.read_csv(flow_file, sep=",")
                date = pd.to_datetime(flow_data[time_name]).values.astype(
                    "datetime64[D]"
                )
                [c, ind1, ind2] = np.intersect1d(
                    date, t_range_list, return_indices=True
                )
                y[k, ind2, j] = flow_data[streamflow_name].values[ind1]
        return y

    def read_relevant_cols(
        self, object_ids=None, t_range_list: list = None, relevant_cols=None, **kwargs
    ) -> Union[np.array, list]:
        """3d data (site_num * time_length * var_num), time-series data"""
        t_range_list = hydro_time.t_range_days(t_range_list)
        nt = t_range_list.shape[0]
        x = np.full([len(object_ids), nt, len(relevant_cols)], np.nan)
        for k in tqdm(range(len(object_ids)), desc="Read forcing data of CAMELS-CC"):
            forcing_file = os.path.join(
                self.data_source_description["CAMELS_FORCING_DIR"],
                object_ids[k] + "_lump_era5_land_forcing.txt",
            )
            forcing_data = pd.read_csv(forcing_file, sep=" ")
            df_date = forcing_data[["Year", "Mnth", "Day"]]
            df_date.columns = ["year", "month", "day"]
            date = pd.to_datetime(df_date).values.astype("datetime64[D]")

            [c, ind1, ind2] = np.intersect1d(date, t_range_list, return_indices=True)
            for j in range(len(relevant_cols)):
                if "evaporation" in relevant_cols[j]:
                    # evaporation value are all negative (maybe upward flux is marked as negative)
                    x[k, ind2, j] = (
                        forcing_data[relevant_cols[j]].values[ind1] * -1 * 1e3
                    )
                # unit of prep and pet is m, tran them to mm
                elif "precipitation" in relevant_cols[j]:
                    prcp = forcing_data[relevant_cols[j]].values
                    # there are a few negative values for prcp, set them 0
                    prcp[prcp < 0] = 0.0
                    x[k, ind2, j] = prcp[ind1] * 1e3
                else:
                    x[k, ind2, j] = forcing_data[relevant_cols[j]].values[ind1]
        return x

    def read_constant_cols(
        self, object_ids=None, constant_cols=None, **kwargs
    ) -> np.array:
        """2d data (site_num * var_num), non-time-series data"""
        raise NotImplementedError(
            "Directly read attributes from nc file by read_attr_xrdataset."
        )

    def get_constant_cols(self) -> np.array:
        """the constant cols in this data_source"""
        data_folder = self.data_source_description["CAMELS_ATTR_DIR"]
        files = np.sort(os.listdir(data_folder))
        attr_types = []
        for file_ in files:
            file = os.path.join(data_folder, file_)
            attr_tmp = pd.read_csv(file, sep=",", dtype={"gage_id": str})
            attr_types = attr_types + attr_tmp.columns[1:].values.tolist()
        return np.array(attr_types)

    def get_relevant_cols(self) -> np.array:
        """the relevant cols in this data_source"""
        forcing_dir = self.data_source_description["CAMELS_FORCING_DIR"]
        forcing_file = os.path.join(forcing_dir, os.listdir(forcing_dir)[0])
        forcing_tmp = pd.read_csv(forcing_file, sep="\s+", dtype={"gage_id": str})
        return forcing_tmp.columns.values[4:]

    def get_target_cols(self) -> np.array:
        """the target cols in this data_source"""
        return np.array(["streamflow"])

    def cache_attributes_xrdataset(self):
        """Convert all the attributes to a single dataframe

        Returns
        -------
        None
        """
        # NOTICE: although it seems that we don't use pint_xarray, we have to import this package
        import pint_xarray  # noqa: F401

        attr_files = Path(self.data_source_description["CAMELS_ATTR_DIR"]).glob("*.csv")
        dataframes = {}
        for file_path in attr_files:
            df = pd.read_csv(file_path, index_col="gage_id")
            dataframes[file_path.stem] = df  # use stem as key

        # merged all data
        merged_data = pd.DataFrame()
        for df in dataframes.values():
            if merged_data.empty:
                merged_data = df
            else:
                merged_data = pd.merge(merged_data, df, on="gage_id", how="outer")

        # Mapping provided units to the variables in the datasets
        units_dict = {
            "p_mean": "mm/day",
            "pet_mean": "mm/day",
            "aet_mean": "mm/day",
            "aridity": "dimensionless",
            "p_seasonality": "dimensionless",
            "high_prec_freq": "days/year",
            "high_prec_dur": "day",
            "high_prec_timing": "dimensionless",
            "low_prec_freq": "days/year",
            "low_prec_dur": "day",
            "low_prec_timing": "dimensionless",
            "frac_snow_daily": "dimensionless",
            # Adding some additional units for other variables based on the provided units dict
            "elev": "m",
            "slope": "m/km",
            # Geology data
            "geol_class_1st": "dimensionless",
            "geol_class_1st_frac": "dimensionless",
            "geol_class_2nd": "dimensionless",
            "geol_class_2nd_frac": "dimensionless",
            "carb_rocks_frac": "dimensionless",
            # Land Cover
            "EvergreenNeedleleafTree": "dimensionless",
            "EvergreenBroadleafTree": "dimensionless",
            # ... other land cover variables ...
            # Permeability and Porosity
            "Permeability": "m^2",
            "Porosity": "dimensionless",
            # Root Depth
            "root_depth_50": "m",
            "root_depth_99": "m",
            # Soil
            "SNDPPT": "percent",
            "CLYPPT": "percent",
            # Topography
            "Length": "m",
            "Area": "km^2",
            "FormFactor": "dimensionless",
            "ShapeFactor": "dimensionless",
            "CompactnessCoefficient": "dimensionless",
            "CirculatoryRatio": "dimensionless",
            "ElongationRatio": "dimensionless",
        }

        # 转换字符串列为分类变量并记录分类映射
        categorical_mappings = {}
        for column in merged_data.columns:
            if merged_data[column].dtype == "object":
                merged_data[column] = merged_data[column].astype("category")
                categorical_mappings[column] = dict(
                    enumerate(merged_data[column].cat.categories)
                )
                merged_data[column] = merged_data[column].cat.codes

        ds = xr.Dataset()
        for column in merged_data.columns:
            attrs = {"units": units_dict.get(column, "unknown")}
            if column in categorical_mappings:
                attrs["category_mapping"] = categorical_mappings[column]

            data_array = xr.DataArray(
                data=merged_data[column].values,
                dims=["basin"],
                # we have set gage_id as index so that it won't be saved as numeric values
                coords={"basin": merged_data.index.values.astype(str)},
                attrs=attrs,
            )
            ds[column] = data_array

        # 将分类映射转换为字符串
        for column in ds.data_vars:
            if "category_mapping" in ds[column].attrs:
                # 将字典转换为字符串
                mapping_str = str(ds[column].attrs["category_mapping"])
                ds[column].attrs["category_mapping"] = mapping_str
        return ds

    def cache_streamflow_xrdataset(self):
        """Save all basins' streamflow data in a netcdf file in the cache directory"""
        variables = self.get_target_cols()
        basins = self.camels_sites["gage_id"].values
        t_range = ["2014-01-01", "2022-01-01"]
        times = [
            hydro_time.t2str(tmp) for tmp in hydro_time.t_range_days(t_range).tolist()
        ]
        data = self.read_target_cols(
            object_ids=basins,
            t_range_list=t_range,
            target_cols=variables,
        )

        return xr.Dataset(
            {
                "streamflow": (
                    ["basin", "time"],
                    data.reshape([len(basins), len(times)])[:, :],
                    {"units": self.streamflow_unit},
                )
            },
            coords={
                "basin": basins,
                "time": pd.to_datetime(times),
            },
        )

    def cache_forcing_xrdataset(self):
        """Save all daymet basin-forcing data in a netcdf file in the cache directory."""
        variables = self.get_relevant_cols()
        basins = self.camels_sites["gage_id"].values
        t_range = ["2014-01-01", "2022-01-01"]
        times = [
            hydro_time.t2str(tmp) for tmp in hydro_time.t_range_days(t_range).tolist()
        ]
        data = self.read_relevant_cols(
            object_ids=basins,
            t_range_list=t_range,
            relevant_cols=variables,
        )
        era5_land_units = {
            "dewpoint_temperature_2m": "K",
            "temperature_2m": "K",
            "skin_temperature": "K",
            "soil_temperature_level_1": "K",
            "soil_temperature_level_2": "K",
            "soil_temperature_level_3": "K",
            "soil_temperature_level_4": "K",
            "lake_bottom_temperature": "K",
            "lake_ice_depth": "m",
            "lake_ice_temperature": "K",
            "lake_mix_layer_depth": "m",
            "lake_shape_factor": "Dimensionless",
            "lake_total_layer_temperature": "K",
            "snow_albedo": "Dimensionless",
            "snow_cover": "%",
            "snow_density": "kg/m^3",
            "snow_depth": "m",
            "snow_depth_water_equivalent": "m",
            "temperature_of_snow_layer": "K",
            "skin_reservoir_content": "m",
            "volumetric_soil_water_layer_1": "Dimensionless",
            "volumetric_soil_water_layer_2": "Dimensionless",
            "volumetric_soil_water_layer_3": "Dimensionless",
            "volumetric_soil_water_layer_4": "Dimensionless",
            "forecast_albedo": "Dimensionless",
            "u_component_of_wind_10m": "m/s",
            "v_component_of_wind_10m": "m/s",
            "surface_pressure": "Pa",
            "leaf_area_index_high_vegetation": "Dimensionless",
            "leaf_area_index_low_vegetation": "Dimensionless",
            "snowfall": "m",
            "snowmelt": "m",
            "surface_latent_heat_flux": "J/m^2",
            "surface_net_solar_radiation": "J/m^2",
            "surface_net_thermal_radiation": "J/m^2",
            "surface_sensible_heat_flux": "J/m^2",
            "surface_solar_radiation_downwards": "J/m^2",
            "surface_thermal_radiation_downwards": "J/m^2",
            # for evaporation and precipitation, we have trans m to mm when reading data
            "evaporation_from_bare_soil": "mm",
            "evaporation_from_open_water_surfaces_excluding_oceans": "mm",
            "evaporation_from_the_top_of_canopy": "mm",
            "evaporation_from_vegetation_transpiration": "mm",
            "potential_evaporation": "mm",
            "runoff": "m",
            "snow_evaporation": "mm",
            "sub_surface_runoff": "m",
            "surface_runoff": "m",
            "total_evaporation": "mm",
            "total_precipitation": "mm",
        }

        return xr.Dataset(
            data_vars={
                **{
                    variables[i]: (
                        ["basin", "time"],
                        data[:, :, i],
                        {"units": era5_land_units[variables[i]]},
                    )
                    for i in range(len(variables))
                }
            },
            coords={
                "basin": basins,
                "time": pd.to_datetime(times),
            },
            attrs={"forcing_type": "era5land"},
        )

    def cache_xrdataset(self):
        """Save all data in a netcdf file in the cache directory"""
        ds_attr = self.cache_attributes_xrdataset()
        ds_attr.to_netcdf(os.path.join(CACHE_DIR, "camelscc_attributes.nc"))
        ds_streamflow = self.cache_streamflow_xrdataset()
        ds_forcing = self.cache_forcing_xrdataset()
        ds = xr.merge([ds_streamflow, ds_forcing])
        ds.to_netcdf(os.path.join(CACHE_DIR, "camelscc_timeseries.nc"))

    def read_ts_xrdataset(
        self,
        gage_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ):
        """read time-series xarray dataset"""
        if var_lst is None:
            return None
        try:
            ts = xr.open_dataset(os.path.join(CACHE_DIR, "camelscc_timeseries.nc"))
        except FileNotFoundError:
            self.cache_xrdataset()
            ts = xr.open_dataset(os.path.join(CACHE_DIR, "camelscc_timeseries.nc"))
        all_vars = ts.data_vars
        if any(var not in ts.variables for var in var_lst):
            raise ValueError(f"var_lst must all be in {all_vars}")
        return ts[var_lst].sel(basin=gage_id_lst, time=slice(t_range[0], t_range[1]))

    def read_attr_xrdataset(self, gage_id_lst=None, var_lst=None, **kwargs):
        """read attribute pandas feather"""
        if var_lst is None or len(var_lst) == 0:
            return None
        try:
            attr = xr.open_dataset(os.path.join(CACHE_DIR, "camelscc_attributes.nc"))
        except FileNotFoundError:
            self.cache_xrdataset()
            attr = xr.open_dataset(os.path.join(CACHE_DIR, "camelscc_attributes.nc"))
        return attr[var_lst].sel(basin=gage_id_lst)

    def read_area(self, gage_id_lst=None):
        """read area of each basin/unit"""
        return self.read_attr_xrdataset(gage_id_lst, ["Area"])

    def read_mean_prcp(self, gage_id_lst=None):
        """read mean precipitation of each basin/unit"""
        return self.read_attr_xrdataset(gage_id_lst, ["p_mean"])


class HydroBasins(HydroData):
    def __init__(self, data_path):
        super().__init__(data_path)

    def get_name(self):
        return " HydroBasins"

    def set_data_source_describe(self):
        self.data_source = "MINIO"
        self.data_source_bucket = "basins-origin"

    def read_BA_xrdataset(self, gage_id_lst: list, var_lst: list, path):
        attr = access_fs.spec_path(path, head="minio")
        if "all" in var_lst:
            return attr.sel(basin=gage_id_lst)
        return attr[var_lst].sel(basin=gage_id_lst)

    def read_MP(self, gage_id_lst, path):
        mean_prep = access_fs.spec_path(path, head="minio")
        return mean_prep["pet_mm_syr"].sel(basin=gage_id_lst)

    def merge_nc_minio_datasets(self, folder_path, basin, var_lst, gap="3h"):
        datasets = []
        basins = []

        file_lst = self.read_file_lst(folder_path)

        for file_path in file_lst:
            basin_id = file_path.split("_")[-1].split(".")[0]
            if basin_id in basin:
                basins.append(basin_id)
                if "ftproot" in file_path:
                    ds = access_fs.spec_path(file_path)
                else:
                    ds = access_fs.spec_path(file_path, head="minio")
                if gap != "1h":
                    ds = self.aggeragate_dataset(ds, gap)
                ds = ds.assign_coords({"basin": basin_id})
                ds = ds.expand_dims("basin")
                datasets.append(ds[var_lst])

        return xr.concat(datasets, dim="basin")

    def aggeragate_dataset(self, ds: xr.Dataset, gap):
        df_res = ds.to_dataframe()
        if "total_evaporation_hourly" in df_res.columns:
            df_res["total_evaporation_hourly"] = (
                df_res["total_evaporation_hourly"].resample(gap, origin="start").sum()
            )
            df_res["total_evaporation_hourly"] *= -1000
            df_res["total_precipitation_hourly"] = (
                df_res["total_precipitation_hourly"].resample(gap, origin="start").sum()
            )
            df_res["total_precipitation_hourly"] *= 1000
        elif "gpm_tp" in df_res.columns:
            df_res["gpm_tp"] = df_res["gpm_tp"].resample(gap, origin="start").sum()
        df_res["streamflow"] = df_res["streamflow"].resample(gap, origin="start").sum()
        df_res = df_res.resample(gap).mean()
        return xr.Dataset.from_dataframe(df_res)

    def read_grid_data(self, file_lst, basin):
        def get_basin_id(file_path):
            return file_path.split("_")[-1].split(".")[0]

        matched_path = next(
            (path for path in file_lst if get_basin_id(path) in basin), None
        )

        if matched_path:
            grid_data = access_fs.spec_path(matched_path, head="minio")
            return grid_data

        return None

    def read_file_lst(self, folder_path):
        if "ftproot" in folder_path:
            return glob.glob(folder_path + "/*")[1:]
        url_path = "s3://" + folder_path
        return conf.FS.glob(url_path + "/**")[1:]



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-22 13:38:07
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-04-28 17:02:31
FilePath: /hydrodatasource/tests/test_streamflow_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''
import pytest
from hydrodatasource.cleaner.streamflow_cleaner import StreamflowCleaner  # 确保引入你的类
import pandas as pd
import matplotlib.pyplot as plt

def test_anomaly_process():
    # 测试径流数据处理功能
    cleaner = StreamflowCleaner("/home/liutianxv1/径流sampledatatest.csv", cwt_row=10)
    # methods默认可以联合调用，也可以单独调用。大多数情况下，默认调用moving_average
    methods = ['wavelet']
    cleaner.anomaly_process(methods)
    print(cleaner.origin_df)
    print(cleaner.processed_df)
    cleaner.processed_df.to_csv("/home/liutianxv1/径流sampledatatest.csv",index=False)

    df = cleaner.processed_df[(cleaner.processed_df['TM']>'2020-07-01')&(cleaner.processed_df['TM']<'2020-10-01')]
    df['TM'] = pd.to_datetime(df['TM'])
    plt.plot(df['TM'], df['Z'], label='Z')  # 可以选择是否添加 marker
    plt.plot(df['TM'], df[str(methods)], label=str(methods))  # 同样，可以选择是否添加 marker
    # 添加图例
    plt.legend()
    # 添加标题和轴标签
    plt.title('TM vs Z and Z_PRoCESS')
    plt.xlabel('TM')
    plt.ylabel('Streamflow')
    # 设置x轴刻度
    xticks = pd.date_range(start=df['TM'].min(), end=df['TM'].max(), freq='15D')
    plt.xticks(xticks, [x.strftime('%Y-%m-%d') for x in xticks], rotation=45)  # 旋转45度，避免标签重叠

    plt.savefig('/home/liutianxv1/plot.png')
    # 显示图表
    plt.show()



"""
Author: Wenyu Ouyang
Date: 2023-01-02 22:23:24
LastEditTime: 2024-03-28 08:34:57
LastEditors: Wenyu Ouyang
Description: read the Global Runoff Data Centre (GRDC) daily data
FilePath: \hydrodata\hydrodatasource\reader\grdc.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
# Global Runoff Data Centre module from ewatercycle: https://github.com/eWaterCycle/ewatercycle/blob/main/src/ewatercycle/observation/grdc.py
import datetime
import os
import pandas as pd
import xarray as xr

from hydrodatasource.downloader.hydrostation import catalogue_grdc
from hydrodatasource.processor.grdc import read_grdc_daily_data


def dailygrdc2netcdf(start_date, end_date, data_dir=None, station_ids=None):
    """
    Parameters
    ----------
    start_date : _type_
        a startDate provided in YYYY-MM-DD
    end_date : _type_
        a endDate provided in YYYY-MM-DD
    """
    nc_file = os.path.join(data_dir, "grdc_daily_data.nc")
    if os.path.exists(nc_file):
        return

    catalogue = catalogue_grdc(data_dir)
    # Create empty lists to store data and metadata
    data_list = []
    meta_list = []

    if station_ids is None:
        # Filter the catalogue based on the provided station IDs
        filenames = os.listdir(data_dir)
        # Extract station IDs from filenames that match the pattern
        station_ids = [
            int(fname.split("_")[0])
            for fname in filenames
            if fname.endswith("_Q_Day.Cmd.txt")
        ]
    catalogue = catalogue[catalogue["grdc_no"].isin(station_ids)]

    # Loop over each station in the catalogue
    for station_id in catalogue["grdc_no"]:
        try:
            st = datetime.datetime.strptime(start_date, "%Y-%m-%d").strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
            et = datetime.datetime.strptime(end_date, "%Y-%m-%d").strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
            df, meta = read_grdc_daily_data(str(station_id), st, et, data_dir)
        except Exception as e:
            print(f"Error reading data for station {station_id}: {e}")
            # Create an empty DataFrame with the same structure
            df = pd.DataFrame(
                columns=["streamflow"],
                index=pd.date_range(start=start_date, end=end_date),
            )
            df["streamflow"] = float("nan")
            meta = {
                "grdc_file_name": "",
                "id_from_grdc": station_id,
                "river_name": "",
                "station_name": "",
                "country_code": "",
                "grdc_latitude_in_arc_degree": float("nan"),
                "grdc_longitude_in_arc_degree": float("nan"),
                "grdc_catchment_area_in_km2": float("nan"),
                "altitude_masl": float("nan"),
                "dataSetContent": "",
                "units": "m³/s",
                "time_series": "",
                "no_of_years": 0,
                "last_update": "",
                "nrMeasurements": "NA",
                "UserStartTime": start_date,
                "UserEndTime": end_date,
                "nrMissingData": 0,
            }

        # Convert the DataFrame to an xarray DataArray and append to the list
        # da = xr.DataArray(
        #     df["streamflow"].values,
        #     coords=[df.index, [station_id]],
        #     dims=["time", "station"],
        # )
        coords_dict = {"time": df.index, "station": [station_id]}
        da = xr.DataArray(
            data=df["streamflow"].values.reshape(-1, 1),
            coords=coords_dict,
            dims=["time", "station"],
            name="streamflow",
            attrs={"units": meta.get("units", "unknown")},
        )
        data_list.append(da)

        # Append metadata
        meta_list.append(meta)

    # Concatenate all DataArrays along the 'station' dimension
    ds = xr.concat(data_list, dim="station")

    # Assign attributes
    ds.attrs["description"] = "Daily river discharge"
    ds.station.attrs["description"] = "GRDC station number"

    # Write the xarray Dataset to a NetCDF file
    ds.to_netcdf(nc_file)

    print("NetCDF file created successfully!")


class GRDCDataHandler:
    def handle(self, configuration):
        aoi_param = configuration["aoi"].aoi_param
        start_time = aoi_param["start_time"]
        end_time = aoi_param["end_time"]
        station_id = aoi_param["station_id"]
        # Based on configuration, read and handle GRDC data specifically
        nc_file = configuration["path"]
        if not os.path.isfile(nc_file):
            dailygrdc2netcdf(start_time, end_time, data_dir=os.path.dirname(nc_file))
        ds = xr.open_dataset(nc_file)
        # choose data for given basin
        return ds.sel(station=int(station_id)).sel(time=slice(start_time, end_time))



#!/home/zhujianfeng/.conda/envs/stac-test/bin/python

import calendar
import datetime
import subprocess
from pathlib import Path

import s3fs
import yaml

from apscheduler.schedulers.blocking import BlockingScheduler  # 后台运行
from minio import Minio


work_dir = Path(__file__).resolve().parent
with open(work_dir / 'privacy_config.yaml', 'r') as f:
    privacy_config = yaml.safe_load(f)
storage_options = {
    'client_kwargs': {'endpoint_url': privacy_config['minio']['client_endpoint']},
    'key': privacy_config['minio']['access_key'],
    'secret': privacy_config['minio']['secret']
}
minioClient = Minio(storage_options['client_kwargs'],
                    storage_options['key'],
                    secret_key=privacy_config['minio']['secret'],
                    secure=False)
fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": privacy_config['minio']['client_endpoint']},
    key=privacy_config['minio']['access_key'],
    secret=privacy_config['minio']['secret']
)

flist = []

sc = BlockingScheduler(timezone="Asia/Shanghai")


@sc.scheduled_job('cron', day_of_week='*', hour='*', minute='55', second='00')
def gpm_downloader():
    # f = open('log.txt', 'a', encoding='utf8')

    date = datetime.datetime.now()
    date = date + datetime.timedelta(hours=-14)
    # date=date.strftime('%Y%m%d')

    year = int(date.strftime('%Y'))
    month = int(date.strftime('%m'))
    day = int(date.strftime('%d'))
    hour = int(date.strftime('%H'))

    new = []
    date = str(year) + str(month).zfill(2) + str(day).zfill(2)

    for time in range(0, 2):
        # time=time*30
        # should.append(date+str(time).zfill(4))

        # url='https://gpm1.gesdisc.eosdis.nasa.gov/daac-bin/OTF/HTTP_services.cgi?FILENAME=%2Fdata%2FGPM_L3%2FGPM_3IMERGHHE.06%2FYYYY%2FAAA%2F3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB-ECCCCCC.EEEE.V06C.HDF5&FORMAT=bmM0Lw&BBOX=3%2C73%2C54%2C136&LABEL=3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB-ECCCCCC.EEEE.V06C.HDF5.SUB.nc4&SHORTNAME=GPM_3IMERGHHE&SERVICE=L34RS_GPM&VERSION=1.02&DATASET_VERSION=06&VARIABLES=precipitationCal%2CrandomError%2CprecipitationUncal%2CIRkalmanFilterWeight%2CHQprecipSource%2CHQprecipitation%2CprobabilityLiquidPrecipitation%2CHQobservationTime%2CIRprecipitation'
        # url='https://gpm1.gesdisc.eosdis.nasa.gov/opendap/GPM_L3/GPM_3IMERGHHE.06/YYYY/AAA/3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB-ECCCCCC.EEEE.V06D.HDF5.nc4?IRkalmanFilterWeight%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,HQprecipSource%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,precipitationCal%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,precipitationUncal%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,HQprecipitation%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,probabilityLiquidPrecipitation%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,HQobservationTime%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,randomError%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,IRprecipitation%5B0:0%5D%5B2530:3159%5D%5B930:1439%5D,time,lon%5B2530:3159%5D,lat%5B930:1439%5D'
        # url='https://gpm1.gesdisc.eosdis.nasa.gov/daac-bin/OTF/HTTP_services.cgi?FILENAME=%2Fdata%2FGPM_L3%2FGPM_3IMERGHHE.06%2F2023%2FAAA%2F3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB-ECCCCCC.EEEE.V06D.HDF5&DATASET_VERSION=06&SERVICE=L34RS_GPM&FORMAT=bmM0Lw&SHORTNAME=GPM_3IMERGHHE&VERSION=1.02&VARIABLES=precipitationCal%2CrandomError%2CprecipitationUncal%2CIRkalmanFilterWeight%2CHQprecipSource%2CHQprecipitation%2CprobabilityLiquidPrecipitation%2CHQobservationTime%2CIRprecipitation&LABEL=3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB-ECCCCCC.EEEE.V06D.HDF5.SUB.nc4&BBOX=3%2C73%2C54%2C136'
        url = ('https://gpm1.gesdisc.eosdis.nasa.gov/daac-bin/OTF/HTTP_services.cgi?'
               'FILENAME=%2Fdata%2FGPM_L3%2FGPM_3IMERGHHE.06%2FYYYY%2FAAA%2F3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB'
               '-ECCCCCC.EEEE.V06E.HDF5&VARIABLES=precipitationCal%2CrandomError%2CprecipitationUncal'
               '%2CIRkalmanFilterWeight%2CHQprecipSource%2CHQprecipitation%2CprobabilityLiquidPrecipitation'
               '%2CHQobservationTime%2CIRprecipitation&DATASET_VERSION=06&BBOX=3%2C73%2C54%2C136&SERVICE=L34RS_GPM'
               '&VERSION=1.02&SHORTNAME=GPM_3IMERGHHE&FORMAT=bmM0Lw&LABEL=3B-HHR-E.MS.MRG.3IMERG.YYYYMMDD-SBBBBBB'
               '-ECCCCCC.EEEE.V06E.HDF5.SUB.nc4')

        YYYY = str(year)
        MM = str(month).zfill(2)
        DD = str(day).zfill(2)

        AAA = str(daynum(year, month, day)).zfill(3)

        EEEE = str(hour * 60 + time * 30).zfill(4)

        start = int(hour * 10000 + time * 30 * 100)
        BBBBBB = str(start).zfill(6)

        end = int(start + 2959)
        CCCCCC = str(end).zfill(6)

        url = url.replace('YYYY', YYYY).replace('MM', MM).replace('DD', DD).replace('AAA', AAA).replace('EEEE',
                                                                                                        EEEE).replace(
            'BBBBBB', BBBBBB).replace('CCCCCC', CCCCCC)
        # print(url)
        # break
        new.append(url)

    with open('./gpm.txt', mode='w') as gpm:
        [gpm.write(str(url) + '\n') for url in new]

    subprocess.run("cat gpm.txt | tr -d '\r' | xargs -n 1 curl -LJO -n -c ~/.urs_cookies -b ~/.urs_cookies", shell=True)

    flist = []
    upload()

    update()

    print(datetime.datetime.now().strftime('%Y%m%d-%H:%M:%S'), '定时任务成功执行')
    # print(datetime.datetime.now().strftime('%Y%m%d-%H:%M:%S'),'定时任务成功执行',file=f)

    # f.close()


def daynum(year, month, day):
    totalday = 0
    if calendar.isleap(year):
        days = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    else:
        days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    for i in range(1, 13):
        if i == month:
            for j in range(1, i):
                totalday = totalday + days[j - 1]
    totalday = totalday + day
    return totalday


def upload():
    for root, dirs, filenames in os.walk('.'):
        for filename in filenames:

            if ".nc4" in filename and "2024" in filename:

                gpm_date = filename[23:31]
                year = gpm_date[0:4]
                month = gpm_date[4:6]
                day = gpm_date[6:8]

                minio_path = f'geodata/gpm/{year}/{month}/{day}/{filename}'

                # print(minio_path)

                # # minio_path = minio_path.replace('_land','')

                try:
                    print(minioClient.fput_object('watermodel-pub', minio_path, os.path.join(root, filename)))
                    flist.append(minio_path)
                    if not os.path.exists(f"/ftproot/geodata/gpm_30m/{year}"):
                        os.mkdir(f"/ftproot/geodata/gpm_30m/{year}")
                    os.system(f"mv {filename} /ftproot/geodata/gpm_30m/{year}")
                except Exception as err:
                    print(err)


import kerchunk.hdf

import os
import ujson
import json

so = dict(mode='rb', storage_options=storage_options, default_fill_cache=False,
          default_cache_type='first')  # args to fs.open()


# default_fill_cache=False avoids caching data in between file chunks to lowers memory usage.

def gen_json(file_url):
    with fs.open(file_url, **so) as infile:
        # print(file_url)
        try:
            h5chunks = kerchunk.hdf.SingleHdf5ToZarr(infile, file_url)
        except:
            f = open('zarrmissed.txt', mode='a')
            print(file_url, file=f)
            f.close()
            return

        outf = file_url[:-4] + '.json'  # file name to save json to
        outf = outf.replace('watermodel-pub', 'test')
        with fs.open(outf, 'wb') as f:
            f.write(ujson.dumps(h5chunks.translate()).encode());


def single_json(nc_list):
    for file in nc_list:
        gen_json(file)


from kerchunk.combine import MultiZarrToZarr


def multi_json(flist):
    # print('multi_json')

    flist.sort()
    lasted = flist[-1]

    filename = lasted.split('/')[-1]

    gpm_date = filename[23:39]
    yyyy = gpm_date[0:4]
    mm = gpm_date[4:6]
    dd = gpm_date[6:8]
    hh = gpm_date[10:12]
    mi = gpm_date[12:14]
    ss = gpm_date[14:16]

    json_list = fs.glob(f"test/geodata/gpm/{yyyy}/{mm}/*/*.json")
    json_list = ['s3://' + str for str in json_list]

    mzz = MultiZarrToZarr(
        json_list,
        target_options=storage_options,
        remote_protocol='s3',
        remote_options=storage_options,
        concat_dims=['time'],
        identical_dims=['lat', 'lon']
    )

    d = mzz.translate()

    # output = f's3://test/geodata/gpm/{yyyy}/{mm}/gpm{yyyy}{mm}_{dd}.json'
    output = f's3://test/geodata/gpm/{yyyy}/{mm}/gpm{yyyy}{mm}_inc.json'
    # days = calendar.monthrange(int(yyyy),int(mm))[1]
    # if int(dd)==days and hh=='23' and mi=='30':
    #     output = f's3://test/geodata/gpm/{yyyy}/{mm}/gpm{yyyy}{mm}_inc.json'
    with fs.open(output, 'wb') as f:
        f.write(ujson.dumps(d).encode())
    print(output, "写入成功！")

    with fs.open('test/geodata/gpm/gpm.json') as f:
        cont = json.load(f)
    cont['end'] = f'{yyyy}-{mm}-{dd}T{hh}:{mi}:00.000000000'
    with fs.open('test/geodata/gpm/gpm.json', 'w') as f:
        json.dump(cont, f)
    print('更新至', filename)


def update():
    nc_list = ['watermodel-pub/' + str for str in flist]
    single_json(nc_list)

    multi_json(flist)


if __name__ == '__main__':

    try:
        sc.start()

    except Exception as e:
        sc.shutdown()
        print(datetime.datetime.now().strftime('%Y%m%d-%H:%M:%S'), '下载任务停止')
        # print(datetime.datetime.now().strftime('%Y%m%d-%H:%M:%S'),'下载任务停止',file=f)



"""
Author: Wenyu Ouyang
Date: 2023-10-13 20:50:01
LastEditTime: 2024-03-28 08:42:28
LastEditors: Wenyu Ouyang
Description: Test downloading function
FilePath: \hydrodata\tests\test_downloader.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import os
import pandas as pd

import hydrodataset as hds

from hydrodatasource.downloader.hydrostation import (
    catalogue_grdc,
    download_grdc_month_data,
    download_grdc_daily_data,
    download_nwis_daily_flow,
)


def test_catalogue_grdc():
    grdc_catalogue = catalogue_grdc(hds.CACHE_DIR)
    assert isinstance(grdc_catalogue, pd.DataFrame)
    assert grdc_catalogue.shape == (10707, 24)
    assert "grdc_no" in grdc_catalogue.columns
    assert "station" in grdc_catalogue.columns
    assert "lat" in grdc_catalogue.columns
    assert "long" in grdc_catalogue.columns
    assert "d_start" in grdc_catalogue.columns
    assert "d_end" in grdc_catalogue.columns
    assert "d_yrs" in grdc_catalogue.columns
    assert "m_start" in grdc_catalogue.columns
    assert "m_end" in grdc_catalogue.columns
    assert "m_yrs" in grdc_catalogue.columns
    assert "t_start" in grdc_catalogue.columns
    assert "t_end" in grdc_catalogue.columns
    assert "t_yrs" in grdc_catalogue.columns
    assert "area" in grdc_catalogue.columns
    assert "altitude" in grdc_catalogue.columns
    assert "r_volume_yr" in grdc_catalogue.columns
    assert "r_height_yr" in grdc_catalogue.columns
    assert "lta_discharge" in grdc_catalogue.columns
    assert "d_miss" in grdc_catalogue.columns
    assert "m_miss" in grdc_catalogue.columns
    assert "wmo_reg" in grdc_catalogue.columns
    assert "sub_reg" in grdc_catalogue.columns


def test_download_grdc_ts(id="2181200"):
    data = download_grdc_month_data(id, hds.CACHE_DIR)

    # Check that the returned data is a dictionary
    assert isinstance(data, dict)

    # Check that the dictionary contains tables with river discharge data
    assert all(isinstance(table, str) for table in data.values())


def test_download_grdc_daily_data(tmp_path):
    station_id = "12345"
    file_path = download_grdc_daily_data(tmp_path, station_id)
    assert os.path.exists(file_path) == False


def test_download_usgs_streamflow():
    camels = hds.Camels()
    sites_id = camels.read_object_ids().tolist()
    date_range = ("2020-10-01", "2021-10-01")
    gage_dict = camels.camels_sites
    save_dir = os.path.join("test_data", "camels_streamflow_2021")
    unit = "cfs"
    qobs = download_nwis_daily_flow(sites_id, date_range, gage_dict, save_dir, unit)
    print(qobs)



"""
该模块用于获取minio服务器中的数据列表，包括：

- `Era5_land`
- `GPM_IMERG_Early`
- `GFS_atmos`
"""

from ..configs.config import MINIO_PARAM
from ..configs.config import FS
import os
import numpy as np
import pandas as pd
import geopandas as gpd
import json


class ERA5LCatalog:
    """
    用于获取era5-land的数据源信息，并搜索minio服务器中的数据范围

    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        datasets (dict): minio服务器中的已有数据集

    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    """

    def __init__(self):
        self._collection_id = "era5-land"
        self._datasources = "ECMWF"
        self._description = "https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-land?tab=overview"
        self._spatialresolution = "0.1 x 0.1; Native resolution is 9 km."
        self._temporalresolution = "hourly"
        self._datasets = self._get_datasets()
        self._bucket_name = "test"

    def _get_datasets(self):
        dss = {}
        ds = {}
        with FS.open(os.path.join(self._bucket_name, "geodata/era5_land/era5l.json")) as f:
            era5 = json.load(f)
            ds["start_time"] = np.datetime64(era5["start"])
            ds["end_time"] = np.datetime64(era5["end"])
            ds["bbox"] = era5["bbox"]
        dss["wis"] = ds
        return dss

    @property
    def collection_id(self):
        return self._collection_id

    @property
    def data_sources(self):
        return self._datasources

    @property
    def description(self):
        return self._description

    @property
    def spatial_resolution(self):
        return self._spatialresolution

    @property
    def temporal_resolution(self):
        return self._temporalresolution

    @property
    def datasets(self):
        return self._datasets

    def search(self, aoi, start_time=None, end_time=None):
        """
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            strt_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        """

        clips = []

        for key, value in self._datasets.items():
            start = start_time
            end = end_time
            if start_time is None:
                start = value["start_time"]
            if end_time is None:
                end = value["end_time"]

            if start < value["start_time"]:
                start = value["start_time"]
            if end > value["end_time"]:
                end = value["end_time"]

            if start <= end:
                df = pd.DataFrame(
                    {
                        "id": [self._collection_id],
                        "dataset": [key],
                        "start_time": [str(start)],
                        "end_time": [str(end)],
                        "geometry": [
                            f"POLYGON(({value['bbox'][0]} {value['bbox'][3]},{value['bbox'][0]} {value['bbox'][1]},\
                                     {value['bbox'][2]} {value['bbox'][1]},{value['bbox'][2]} {value['bbox'][3]},{value['bbox'][0]} {value['bbox'][3]}))"
                        ],
                    }
                )
                df["geometry"] = gpd.GeoSeries.from_wkt(df["geometry"])
                gdf = gpd.GeoDataFrame(df, geometry="geometry", crs="EPSG:4326")

                clips.append(gdf.clip(aoi))

        return pd.concat(clips)


class GPMCatalog:
    """
    用于获取gpm的数据源信息，并搜索minio服务器中的数据范围

    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        datasets (dict): minio服务器中的已有数据集

    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    """

    def __init__(self):
        self._collection_id = "gpm-imerg-early"
        self._datasources = "NASA & JAXA"
        self._description = (
            "https://disc.gsfc.nasa.gov/datasets/GPM_3IMERGHHE_06/summary"
        )
        self._spatialresolution = "0.1 x 0.1; Native resolution is 9 km. (60°S-60°N)"
        self._temporalresolution = "half-hourly; 1 day"
        self._bucket_name = "test"
        self._datasets = self._get_datasets()

    def _get_datasets(self):
        dss = {}
        lds = []
        ds = {}
        with FS.open(os.path.join(self._bucket_name, "geodata/gpm/gpm.json")) as f:
            gpm = json.load(f)
            ds["time_resolution"] = "30 minutes"
            ds["start_time"] = np.datetime64(gpm["start"])
            ds["end_time"] = np.datetime64(gpm["end"])
            ds["bbox"] = gpm["bbox"]
        lds.append(ds)

        ds = {}
        with FS.open(os.path.join(self._bucket_name, "geodata/gpm1d/gpm1d.json")) as f:
            gpm = json.load(f)
            ds["time_resolution"] = "1 day"
            ds["start_time"] = np.datetime64(gpm["start"])
            ds["end_time"] = np.datetime64(gpm["end"])
            ds["bbox"] = gpm["bbox"]
        lds.append(ds)
        dss["wis"] = lds

        lds = []
        ds = {}
        with FS.open(os.path.join(self._bucket_name, "camdata/gpm/gpm.json")) as f:
            gpm = json.load(f)
            ds["time_resolution"] = "30 minutes"
            ds["start_time"] = np.datetime64(gpm["start"])
            ds["end_time"] = np.datetime64(gpm["end"])
            ds["bbox"] = gpm["bbox"]
        lds.append(ds)

        ds = {}
        with FS.open(os.path.join(self._bucket_name, "camdata/gpm1d/gpm1d.json")) as f:
            gpm = json.load(f)
            ds["time_resolution"] = "1 day"
            ds["start_time"] = np.datetime64(gpm["start"])
            ds["end_time"] = np.datetime64(gpm["end"])
            ds["bbox"] = gpm["bbox"]
        lds.append(ds)
        dss["camels"] = lds

        return dss

    @property
    def collection_id(self):
        return self._collection_id

    @property
    def data_sources(self):
        return self._datasources

    @property
    def description(self):
        return self._description

    @property
    def spatial_resolution(self):
        return self._spatialresolution

    @property
    def temporal_resolution(self):
        return self._temporalresolution

    @property
    def datasets(self):
        return self._datasets

    def search(self, aoi, start_time=None, end_time=None):
        """
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            strt_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        """

        clips = []

        for key, value in self._datasets.items():
            for v in value:
                start = start_time
                end = end_time
                if start_time is None:
                    start = v["start_time"]
                if end_time is None:
                    end = v["end_time"]

                if start < v["start_time"]:
                    start = v["start_time"]
                if end > v["end_time"]:
                    end = v["end_time"]

                if start <= end:
                    df = pd.DataFrame(
                        {
                            "id": [self._collection_id],
                            "dataset": [key],
                            "time_resolution": v["time_resolution"],
                            "start_time": [str(start)],
                            "end_time": [str(end)],
                            "geometry": [
                                f"POLYGON(({v['bbox'][0]} {v['bbox'][3]},{v['bbox'][0]} {v['bbox'][1]},\
                                         {v['bbox'][2]} {v['bbox'][1]},{v['bbox'][2]} {v['bbox'][3]},{v['bbox'][0]} {v['bbox'][3]}))"
                            ],
                        }
                    )
                    df["geometry"] = gpd.GeoSeries.from_wkt(df["geometry"])
                    gdf = gpd.GeoDataFrame(df, geometry="geometry", crs="EPSG:4326")

                    clips.append(gdf.clip(aoi))

        return pd.concat(clips)


class GFSCatalog:
    """
    用于获取gfs的数据源信息，并搜索minio服务器中的数据范围

    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        datasets (dict): minio服务器中的已有数据集

    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    """

    def __init__(self, variable="tp"):
        self._variable = variable
        self._collection_id = f"gfs_atmos.{variable}"
        self._datasources = "NOAA"
        self._description = (
            "https://www.emc.ncep.noaa.gov/emc/pages/numerical_forecast_systems/gfs.php"
        )
        self._spatialresolution = "0.25 x 0.25"
        self._temporalresolution = "hourly; 1-120h"
        self._bucket_name = "test"
        self._datasets = self._get_datasets()

    def _get_datasets(self):
        dss = {}

        ds = {}
        with FS.open(os.path.join(self._bucket_name, "geodata/gfs/gfs.json")) as f:
            gfs = json.load(f)
        dss["wis"] = gfs[self._variable]

        return dss

    @property
    def variable(self):
        return self._variable

    @property
    def collection_id(self):
        return self._collection_id

    @property
    def data_sources(self):
        return self._datasources

    @property
    def description(self):
        return self._description

    @property
    def spatial_resolution(self):
        return self._spatialresolution

    @property
    def temporal_resolution(self):
        return self._temporalresolution

    @property
    def datasets(self):
        return self._datasets

    def search(self, aoi, start_time=None, end_time=None):
        """
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            start_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        """

        clips = []

        for key, value in self._datasets.items():
            for v in value:
                start = start_time
                end = end_time
                if start_time is None:
                    start = np.datetime64(v["start"])
                if end_time is None:
                    end = np.datetime64(v["end"])

                if start < np.datetime64(v["start"]):
                    start = np.datetime64(v["start"])

                if end > np.datetime64(v["end"]):
                    end = np.datetime64(v["end"])

                if start <= end:
                    df = pd.DataFrame(
                        {
                            "id": [self._collection_id],
                            "dataset": [key],
                            "start_time": [str(start)],
                            "end_time": [str(end)],
                            "geometry": [
                                f"POLYGON(({v['bbox'][0]} {v['bbox'][3]},{v['bbox'][0]} {v['bbox'][1]},{v['bbox'][2]} {v['bbox'][1]},{v['bbox'][2]} {v['bbox'][3]},{v['bbox'][0]} {v['bbox'][3]}))"
                            ],
                        }
                    )
                    df["geometry"] = gpd.GeoSeries.from_wkt(df["geometry"])
                    gdf = gpd.GeoDataFrame(df, geometry="geometry", crs="EPSG:4326")

                    clips.append(gdf.clip(aoi))

        return pd.concat(clips)



import requests as r
import datetime
import numpy as np

from ..utils.utils import validate


class LandsatCatalog:
    def __init__(self):
        self._root = (
            "https://landsatlook.usgs.gov/stac-server"  # Landsat STAC API Endpoint
        )

        stac_response = r.get(self._root).json()
        catalog_links = stac_response["links"]
        self._search = [l["href"] for l in catalog_links if l["rel"] == "search"][
            0
        ]  # retreive search endpoint from STAC Catalogs

    @property
    def root(self):
        return self._root

    def set_root(self, url):
        self._root = url

        stac_response = r.get(self._root).json()
        catalog_links = stac_response["links"]
        self._search = [l["href"] for l in catalog_links if l["rel"] == "search"][
            0
        ]  # retreive search endpoint from STAC Catalogs

    def search(
        self,
        start_date: str = "1972-01-01",
        end_date: str = None,
        aoi=None,
        bbox=None,
        limit=1000,
    ):
        start_date = validate(start_date, "%Y-%m-%d", "start_date格式错误，应为yyyy-mm-dd")
        if end_date is None:
            end_date = datetime.date.today().strftime("%Y-%m-%d")
        end_date = validate(end_date, "%Y-%m-%d", "end_date格式错误，应为yyyy-mm-dd")

        if start_date > end_date:
            raise ValueError("开始日期不能大于结束日期！")

        params = {}
        params["limit"] = limit

        if aoi is None and bbox is None:
            raise ValueError("需指定aoi或bbox的范围！")
        elif bbox is None:
            params["intersects"] = aoi
        else:
            params["bbox"] = bbox

        landsat_dataset = []

        # dt = f"{start_date.strftime("%Y-%m-%d")}T00:00:00Z/{end_date.strftime("%Y-%m-%d")}T23:59:59Z"
        for y in range(end_date.year, start_date.year - 1, -1):
            if y == start_date.year:
                dt = start_date.strftime("%Y-%m-%d") + "T00:00:00Z"
            else:
                dt = f"{str(y)}-01-01T00:00:00Z"

            if y == end_date.year:
                dt = dt + "/" + end_date.strftime("%Y-%m-%d") + "T23:59:59Z"
            else:
                dt = dt + f"/{str(y)}-12-31T23:59:59Z"

            params["datetime"] = dt

            query = r.post(
                self._search,
                json=params,
            ).json()
            try:
                landsat_dataset += query["features"]
            except:
                raise Exception(query)

        return landsat_dataset

    def get_hrefs(self, ids):
        hrefs = []
        params = {}
        params["ids"] = ids

        query = r.post(
            self._search,
            json=params,
        ).json()
        for feature in query["features"]:
            feature_hrefs = []
            for key in feature["assets"].keys():
                if "data" in feature["assets"][key]["roles"]:
                    feature_hrefs.append(feature["assets"][key]["href"])
            hrefs.append(feature_hrefs)

        return hrefs


from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session


class SentinelCatalog:
    def __init__(self):
        self._root = "https://sh.dataspace.copernicus.eu/api/v1/catalog/1.0.0/"

        stac_response = r.get(self._root).json()
        catalog_links = stac_response["links"]
        self._search = [l["href"] for l in catalog_links if l["rel"] == "search"][
            0
        ]  # retreive search endpoint from STAC Catalogs

        self._oauth = None

    @property
    def root(self):
        return self._root

    @property
    def token(self):
        return self._oauth

    def set_root(self, url):
        self._root = url

        stac_response = r.get(self._root).json()
        catalog_links = stac_response["links"]
        self._search = [l["href"] for l in catalog_links if l["rel"] == "search"][
            0
        ]  # retreive search endpoint from STAC Catalogs

    def get_token(self, client_id=None, client_secret=None):
        if client_id is not None:
            self.__client_id = client_id
        if client_secret is not None:
            self.__client_secret = client_secret
        # Your client credentials
        # self.__client_id = '9bff23bb-227a-4a35-9b36-22ab23b9dfaf'
        # self.__client_secret = '7Ea%6-CM%hG,k![.DQB?i-bw&Y*Q.eG6c:_{uyUW'

        if self.__client_id is None or self.__client_secret is None:
            raise ValueError("认证信息错误！")

        # Create a session
        client = BackendApplicationClient(client_id=self.__client_id)
        oauth = OAuth2Session(client=client)

        # Get token for the session
        token = oauth.fetch_token(
            token_url="https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
            client_secret=client_secret,
        )

        # All requests using this session will have an access token automatically added
        # resp = oauth.get("https://services.sentinel-hub.com/oauth/tokeninfo")
        # print(resp.content)
        def sentinelhub_compliance_hook(response):
            response.raise_for_status()
            return response

        oauth.register_compliance_hook(
            "access_token_response", sentinelhub_compliance_hook
        )

        self._oauth = oauth

    def search(
        self,
        start_date: str = "2013-01-01",
        end_date: str = None,
        collections=["sentinel-1-grd", "sentinel-2-l1c", "sentinel-2-l2a"],
        aoi=None,
        bbox=None,
        limit=100,
    ):
        if self._oauth is None:
            raise ValueError("请验证token！")

        start_date = validate(start_date, "%Y-%m-%d", "start_date格式错误，应为yyyy-mm-dd")
        if end_date is None:
            end_date = datetime.date.today().strftime("%Y-%m-%d")
        end_date = validate(end_date, "%Y-%m-%d", "end_date格式错误，应为yyyy-mm-dd")

        if start_date > end_date:
            raise ValueError("开始日期不能大于结束日期！")

        params = {}
        params["limit"] = limit

        if aoi is None and bbox is None:
            raise ValueError("需指定aoi或bbox的范围！")
        elif bbox is None:
            params["intersects"] = aoi
        else:
            params["bbox"] = bbox

        sentinel_dataset = []

        d = start_date
        delta = datetime.timedelta(days=1)
        while d <= end_date:
            dt = d.strftime("%Y-%m-%d")
            params["datetime"] = f"{dt}T00:00:00Z/{dt}T23:59:59Z"

            for col in collections:
                params["collections"] = [col]

                try:
                    query = self._oauth.post(
                        self._search,
                        json=params,
                    ).json()
                    sentinel_dataset += query["features"]
                except:
                    raise Exception(query)
                    # print(query)

                d += delta

        return sentinel_dataset

    def get_hrefs(self, ids):
        hrefs = []
        params = {}
        params["ids"] = ids

        query = self._oauth.post(
            self._search,
            json=params,
        ).json()
        for feature in query["features"]:
            feature_hrefs = []
            for key in feature["assets"].keys():
                if "data" in feature["assets"][key]["roles"]:
                    feature_hrefs.append(feature["assets"][key]["href"])
            hrefs.append(feature_hrefs)

        return hrefs



from hydrodatasource.reader.postgres import read_data


def test_read_rainfall_data():
    data = read_data(stcd='10800800', datatype='rain', start_time='2013-08-16T08:00:00', end_time='2013-08-17T08:00:00')
    print(data)


def test_read_streamflow_data():
    data = read_data(stcd='21401550', datatype='streamflow', start_time='2020-07-01T00:00:00',
                     end_time='2020-07-31T23:00:00')
    print(data)



"""
该模块用于alos-dem数据下载

数据源: [Microsoft Planetary Computer](https://planetarycomputer.microsoft.com/dataset/alos-dem)

数据目录: [https://planetarycomputer.microsoft.com/api/stac/v1/](https://planetarycomputer.microsoft.com/api/stac/v1/)

数据说明: [https://www.eorc.jaxa.jp/ALOS/en/aw3d30/aw3d30v3.2_product_e_e1.2.pdf](https://www.eorc.jaxa.jp/ALOS/en/aw3d30/aw3d30v3.2_product_e_e1.2.pdf)

"""

import os

# from pystac_client import Client
import requests

from .downloader import download_sigletasking


class Alos_DEM:
    def __init__(self):
        self._sources = {
            "microsoft": {
                "url": "https://planetarycomputer.microsoft.com/api/stac/v1/",
                "collection_name": "alos-dem",
            },
        }
        self._hrefs = []

    def list_sources(self):
        """
        返回stac列表

        Returns:
            sources (list): stac列表
        """
        return self._sources

    def add_source(self, source_name, url, collection_name):
        """
        添加自定义stac内容

        Args:
            source_name (str): 源名称
            url (list): STAC根路径地址
            collection_name (str): 数据集名称

        Returns:
            source (dict): 符合条件的url地址列表
        """

        new = {source_name: {"url": url, "collection_name": collection_name}}
        self._sources[source_name] = {"url": url, "collection_name": collection_name}

        return new

    def search(self, source="microsoft", bbox=None, intersects=None):
        """
        搜索符合条件的dem，并返回url地址列表

        Args:
            catalog_url (str): STAC根路径地址，默认为[微软planetarycomputer](https://planetarycomputer.microsoft.com)
            bbox (list): 搜索的矩形范围
            intersects (str): 搜索范围，应符合GeoJSON的geometry属性格式

        Returns:
            hrefs (list): 符合条件的url地址列表
        """

        if bbox is None:
            bbox = [115, 38, 136, 54]
        if source not in self._sources.keys():
            return

        stac = self._sources[source]["url"]
        stac_response = requests.get(stac).json()
        catalog_links = stac_response["links"]
        search = [l["href"] for l in catalog_links if l["rel"] == "search"][0]

        params = {"collections": [self._sources[source]["collection_name"]]}
        if bbox is not None:
            params["bbox"] = bbox
        if intersects is not None:
            params["intersects"] = intersects

        query = requests.post(search, json=params).json()

        features = query["features"]

        for feat in features:
            self._hrefs.append(feat["assets"]["data"]["href"])

        return self._hrefs

    def download(self, save_dir=".", cover=False):
        """
        下载列表中的dem数据

        Args:
            save_dir (str): 本地保存目录
            cover (bool): 若文件已存在，是否覆盖

        """

        if self._hrefs is not None:
            for href in self._hrefs:
                url = requests.utils.urlparse(href)
                file_path = os.path.join(save_dir, url.path.split("/")[-1])

                if os.path.exists(file_path) and not cover:
                    continue

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                download_sigletasking(href, file_path)



"""
Author: Wenyu Ouyang
Date: 2023-10-14 11:51:22
LastEditTime: 2024-03-28 08:38:29
LastEditors: Wenyu Ouyang
Description: Reading streamflow/water level data from hydrostation data source
FilePath: \hydrodata\hydrodatasource\downloader\hydrostation.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
from datetime import datetime
import numpy as np
import pandas as pd
import os
import warnings
import pytz
import requests
from dataretrieval import nwis
from pygeohydro import NWIS

from hydrodatasource.downloader.downloader import (
    download_ftp_file,
    unzip_file,
    wget_download,
)


def catalogue_grdc(save_dir):
    # URL for the GRDC catalogue
    file_url = "ftp://ftp.bafg.de/pub/REFERATE/GRDC/catalogue/grdc_stations.zip"

    # Download the zip file using FTP
    save_path = wget_download(file_url, save_dir)
    # Extract the Excel file from the zip
    unzip_folder = unzip_file(save_path)
    # Read the Excel file
    station_file = os.path.join(unzip_folder, "GRDC_Stations.xlsx")
    grdc_catalogue = pd.read_excel(station_file, sheet_name="station_catalogue")

    # Cleanup the data
    grdc_catalogue.replace(["n.a.", "-999.0"], np.nan, inplace=True)

    # Convert certain columns to integer or numeric data types
    cols_to_int = [
        "wmo_reg",
        "sub_reg",
        "d_start",
        "d_end",
        "d_yrs",
        "m_start",
        "m_end",
        "m_yrs",
        "t_start",
        "t_end",
        "t_yrs",
    ]
    grdc_catalogue[cols_to_int] = grdc_catalogue[cols_to_int].astype(pd.Int64Dtype)

    cols_to_numeric = [
        "lat",
        "long",
        "area",
        "altitude",
        "d_miss",
        "m_miss",
        "lta_discharge",
        "r_volume_yr",
        "r_height_yr",
    ]
    grdc_catalogue[cols_to_numeric] = grdc_catalogue[cols_to_numeric].astype(float)

    return grdc_catalogue


def download_grdc_month_data(id, save_dir):
    """
    Interface with the Global Runoff Data Centre database of Monthly Time Series.

    For daily data, you have to manually apply using your email in the following link:
    https://portal.grdc.bafg.de/applications/public.html?publicuser=PublicUser#dataDownload/Stations

    Parameters
    -----------
    id (str)
        Identifier for a station. It's called "grdc no" in the catalogue.
                it will be transformed to int type.

    Returns
    -------
    dict
        A dictionary containing tables file path.
    """

    # catalogue_grdc to get the catalogue.
    catalogue = catalogue_grdc(save_dir)

    # Filter the catalogue based on the provided id
    catalogue_filtered = catalogue[catalogue["grdc_no"] == int(id)]

    # Retrieve the WMO region from the filtered catalogue
    wmo_region = catalogue_filtered["wmo_reg"].iloc[0]

    # Load the grdcLTMMD.csv file
    this_dir = os.path.dirname(os.path.abspath(__file__))
    grdc_ltmmd_file = os.path.join(this_dir, "grdcLTMMD.csv")
    grdc_ltmmd = pd.read_csv(grdc_ltmmd_file)
    # Retrieve ftp server location based on the WMO region
    zip_file_url = grdc_ltmmd.loc[
        grdc_ltmmd["WMO_Region"] == wmo_region, "Archive"
    ].values[0]

    # Download the ZIP file from FTP server
    file_path = download_ftp_file(zip_file_url, save_dir)

    data_dir = unzip_file(file_path)

    # Table names
    tablenames = ["LTVD", "LTVM", "PVD", "PVM", "YVD", "YVM"]

    # Assuming 'id' is already defined
    fname = [f"{id}_Q_{table}.csv" for table in tablenames]

    # Initialise empty dictionary of tables
    tables = {}

    # Populate tables
    for j, fpath in enumerate(fname):
        full_path = os.path.join(data_dir, fpath)
        if os.path.exists(full_path):
            tables[tablenames[j]] = full_path
        else:
            print(f"{tablenames[j]} data are not available at this station")
            tables[tablenames[j]] = None
    return tables


def download_grdc_daily_data(station_id, data_dir):
    file_path = os.path.join(data_dir, f"{station_id}_Q_Day.Cmd.txt")
    if not os.path.exists(file_path):
        message = (
            "For GRDC stations, easily use table-view to see; "
            "Then, when you download the data, you can select all the station ids you need to download the data. "
            "Remember to choose the export format as 'GRDC Export Format (daily data only)'."
        )
        warnings.warn(message, UserWarning)
    return file_path


def download_nwis(
    gage_id_file="path_to_gauge_information.txt",
    start_date="2016-01-01",
    end_date="2023-06-30",
    param_code="00065",
):
    # TODO: merge this func with get_nwis_stream_data
    with open(gage_id_file, "r", encoding="utf-8") as file:
        content = file.readlines()

    gage_ids = [line.split("\t")[1].strip() for line in content[1:] if line.strip()]
    # sites = gage_ids

    for gage_id in gage_ids:
        # get instantaneous values (iv)
        df, md = nwis.get_iv(
            sites=gage_id, start=start_date, end=end_date, parameterCd=param_code
        )

        df.to_csv(
            f"{str(gage_id)}.txt",
            sep=" ",
            header=False,
            na_rep="",
            float_format="%6.2f",
        )

        # 打开原文件，并读取所有内容
        with open(f"{str(gage_id)}.txt", "r", encoding="utf-8") as file:
            content = file.read()

        # 使用replace方法去除所有的双引号
        content_no_quotes = content.replace('"', "")

        # 写入新的文件
        with open(f"{str(gage_id)}.txt", "w", encoding="utf-8") as file:
            file.write(content_no_quotes)


def get_nwis_stream_data(
    save_dir=".",
    sites=None,
    start=None,
    end=None,
    multi_index=True,
    ssl_check=True,
    param_code="All",
    standardlize=True,
):
    """
    用来获取usgs相关流域的数据

    Args:
        sites:站点id
        start:起始时间
        end:结束时间
        multi_index:多索引
        ssl_check:检查ssl
        parameterCd:用于获取部分数据，如果不指定这个参数，则获取全部数据
            00060代表流量
            00065代表水位
            出现其余数据类型，请参考USGS官网或nwis github
        standardlize:用于规范化存储txt

    Return:
        (type:DataFrame)
        datetime:时间
        site_no:站点ID
        00060:同上
        00065：同上
        00060_cd：测量情况，A代表实测，A,e代表经过处理，M代表无数据
        00065_cd: 同上
    """

    df, md = nwis.get_iv(sites=sites, start=start, end=end, parameterCd=param_code)
    df.to_csv(
        save_dir + sites + ".txt", sep=" ", header=True, na_rep="", float_format="%6.2f"
    )

    if standardlize == True:
        # 打开原文件，并读取所有内容
        with open(save_dir + sites + ".txt", "r", encoding="utf-8") as file:
            content = file.read()

        # 使用replace方法去除所有的双引号
        content_no_quotes = content.replace('"', "")

        # 写入新的文件
        with open(save_dir + sites + ".txt", "w", encoding="utf-8") as file:
            file.write(content_no_quotes)
    return [df, md]


# 等项目忙完再补充，这个是获得所有数据的
def get_nwis_full_data(
    save_dir="",
    sites=None,
    start=None,
    end=None,
    multi_index=True,
    wide_format=True,
    datetime_index=True,
    state=None,
    service="iv",
    ssl_check=True,
    **kwargs,
):
    """
    完整的获取各种nwis数据的方法

    参考：https://github.com/DOI-USGS/dataretrieval-python/blob/master/dataretrieval/nwis.py#L1289

    Args:
        save_dir = '':保存目录
        sites:站点的ID，例如:'01013500'
        start:开始时间，如start='2016-01-01'
        end:结束时间，如end='2016-01-02'
        multi_index:多标签,If False, a dataframe with a single-level index (datetime) is returned
        wide_format:If True, return data in wide format with multiple samples per row and one row per time
        service:索取数据类型
            - 'iv' : instantaneous data，即时数据，15分钟尺度
            - 'dv' : daily mean data，日尺度
            - 'qwdata' : discrete samples
            - 'site' : 站点描述
            - 'measurements' : 流量测量值
            - 'peaks': 流量峰值
            - 'gwlevels': 地表水等级
            - 'pmcodes': 获取参数代码
            - 'water_use': 使用水数据
            - 'ratings': get rating table
            - 'stat': 获取数据，A代表实测，A.e代表经过处理，M代表无
        ssl_check:检查ssl
        **kwargs:可调用的其余参数
    """
    pass


def download_nwis_daily_flow(
    usgs_site_ids: list,
    date_tuple: tuple,
    gage_dict: dict,
    save_dir: str,
    unit: str = "cfs",
) -> pd.DataFrame:
    """
    Download USGS flow data by HyRivers' pygeohydro tool.
    The tool's tutorial: https://github.com/cheginit/HyRiver-examples/blob/main/notebooks/nwis.ipynb

    Parameters
    ----------
    usgs_site_ids
        ids of USGS sites
    date_tuple
        start and end date
    gage_dict
        a dict containing gage's ids and the correspond HUC02 ids
    save_dir
        where we save streamflow data in files like CAMELS
    unit
        unit of streamflow, cms or cfs

    Returns
    -------
    pd.DataFrame
        streamflow data -- index is date; column is gage id
    """

    nwis = NWIS()
    qobs = nwis.get_streamflow(usgs_site_ids, date_tuple, mmd=False)
    # the unit of qobs is cms, but unit in CAMELS and GAGES-II is cfs, so here we transform it
    # use round(2) because in both CAMELS and GAGES-II, data with cfs only have two float digits
    if unit == "cfs":
        qobs = (qobs * 35.314666212661).round(2)
    dates = qobs.index
    camels_format_index = ["GAGE_ID", "Year", "Mnth", "Day", f"streamflow({unit})"]
    year_month_day = pd.DataFrame(
        [[dt.year, dt.month, dt.day] for dt in dates], columns=camels_format_index[1:4]
    )
    if "STAID" in gage_dict:
        gage_id_key = "STAID"
    elif "gauge_id" in gage_dict:
        gage_id_key = "gauge_id"
    elif "gage_id" in gage_dict:
        gage_id_key = "gage_id"
    else:
        raise NotImplementedError("No such gage id name")
    if "HUC02" in gage_dict:
        huc02_key = "HUC02"
    elif "huc_02" in gage_dict:
        huc02_key = "huc_02"
    else:
        raise NotImplementedError("No such huc02 id")
    read_sites = [col[5:] for col in qobs.columns.values]
    for site_id in usgs_site_ids:
        if site_id not in read_sites:
            df_flow = pd.DataFrame(
                np.full(qobs.shape[0], np.nan), columns=camels_format_index[4:5]
            )
        else:
            qobs_i = qobs["USGS-" + site_id]
            df_flow = pd.DataFrame(qobs_i.values, columns=camels_format_index[4:5])
        df_id = pd.DataFrame(
            np.full(qobs.shape[0], site_id), columns=camels_format_index[:1]
        )
        new_data_df = pd.concat([df_id, year_month_day, df_flow], axis=1)
        # output the result
        i_basin = gage_dict[gage_id_key].values.tolist().index(site_id)
        huc_id = gage_dict[huc02_key][i_basin]
        output_huc_dir = os.path.join(save_dir, huc_id)
        if not os.path.isdir(output_huc_dir):
            os.makedirs(output_huc_dir)
        output_file = os.path.join(output_huc_dir, site_id + "_streamflow_qc.txt")
        if os.path.isfile(output_file):
            os.remove(output_file)
        new_data_df.to_csv(
            output_file, header=True, index=False, sep=",", float_format="%.2f"
        )
    return qobs


def _usgsflow_df_label(usgs_text: str) -> str:
    usgs_text = usgs_text.replace(",", "")
    if usgs_text == "Discharge":
        return "cfs"
    elif usgs_text == "Gage":
        return "height"
    else:
        return usgs_text


def _process_usgs_response_text(file_name: str):
    extractive_params = {}
    with open(file_name, "r") as f:
        lines = f.readlines()
        i = 0
        params = False
        while "#" in lines[i]:
            # TODO figure out getting height and discharge code efficently
            the_split_line = lines[i].split()[1:]
            if params:
                print(the_split_line)
                if len(the_split_line) < 2:
                    params = False
                else:
                    extractive_params[
                        the_split_line[0] + "_" + the_split_line[1]
                    ] = _usgsflow_df_label(the_split_line[2])
            if len(the_split_line) > 2 and the_split_line[0] == "TS":
                params = True
            i += 1
        with open(file_name.split(".")[0] + "data.tsv", "w") as t:
            t.write("".join(lines[i:]))
        return file_name.split(".")[0] + "data.tsv", extractive_params


def _create_csv_for_usgsflow(file_path: str, params_names: dict, site_number: str):
    """
    Function that creates the final version of the CSV files

    Parameters
    ----------
    file_path : str
        [description]
    params_names : dict
        [description]
    site_number : str
        [description]
    """
    df = pd.read_csv(file_path, sep="\t")
    for key, value in params_names.items():
        df[value] = df[key]
    df.to_csv(f"{site_number}_flow_data.csv")


def download_usgs_data(
    start_date: datetime, end_date: datetime, site_number: str
) -> pd.DataFrame:
    """This method could also be used to download usgs streamflow data"""
    # TODO: to be merged with get_nwis_stream_data; may be useful when handling with hourly data
    base_url = (
        "https://nwis.waterdata.usgs.gov/usa/nwis/uv/?cb_00060=on&cb_00065&format=rdb&"
    )
    full_url = (
        base_url
        + "site_no="
        + site_number
        + "&period=&begin_date="
        + start_date.strftime("%Y-%m-%d")
        + "&end_date="
        + end_date.strftime("%Y-%m-%d")
    )
    print("Getting request from USGS")
    print(full_url)
    r = requests.get(full_url)
    with open(f"{site_number}.txt", "w") as f:
        f.write(r.text)
    print("Request finished")
    response_data = _process_usgs_response_text(f"{site_number}.txt")
    _create_csv_for_usgsflow(response_data[0], response_data[1], site_number)
    return pd.read_csv(f"{site_number}_flow_data.csv")


def _get_usgsflow_timezone_map():
    return {
        "EST": "America/New_York",
        "EDT": "America/New_York",
        "CST": "America/Chicago",
        "CDT": "America/Chicago",
        "MDT": "America/Denver",
        "MST": "America/Denver",
        "PST": "America/Los_Angeles",
        "PDT": "America/Los_Angeles",
    }


def process_intermediate_csv(df: pd.DataFrame):
    # Remove garbage first row
    # TODO: check if more rows are garbage
    df = df.iloc[1:]
    time_zone = df["tz_cd"].iloc[0]
    time_zone = _get_usgsflow_timezone_map()[time_zone]
    old_timezone = pytz.timezone(time_zone)
    new_timezone = pytz.timezone("UTC")
    # This assumes timezones are consistent throughout the USGS stream (this should be true)
    df["datetime"] = df["datetime"].map(
        lambda x: old_timezone.localize(
            datetime.strptime(x, "%Y-%m-%d %H:%M")
        ).astimezone(new_timezone)
    )
    df["cfs"] = pd.to_numeric(df["cfs"], errors="coerce")
    max_flow = df["cfs"].max()
    min_flow = df["cfs"].min()
    count_nan = len(df["cfs"]) - df["cfs"].count()
    print(f"there are {count_nan} nan values")
    return df[df.datetime.dt.minute == 0], max_flow, min_flow



import os

import geopandas as gpd
import h5py
import numpy as np
import pandas as pd
import xarray as xr
from pandas import DataFrame
from pandas.core.indexes.api import default_index
import hydrodatasource.processor.mask as hpm
import hydrodatasource.configs.config as hdscc
from hydrodatasource.processor.basin_mean_rainfall import rainfall_average
from hydrodatasource.reader import access_fs


def query_path_from_metadata(basin_id, time_start=None, time_end=None, bbox=None, data_source='gpm'):
    # query path from other columns from metadata.csv
    metadata_df = pd.read_csv(f's3://grids-origin/{data_source}_metadata.csv', storage_options=hdscc.MINIO_PARAM)
    source_list = [x for x in metadata_df['path'] if ((data_source in x) | (data_source.upper() in x))]
    paths = metadata_df[metadata_df['path'].isin(source_list)]
    if time_start is not None:
        paths = paths[paths['time_start'] >= time_start]
    if time_end is not None:
        paths = paths[paths['time_end'] <= time_end]
    if (data_source == 'gpm') or (data_source == 'smap') or (data_source == 'era5_land') or (data_source == 'era5'):
        if bbox is not None:
            paths = paths[
                (paths['bbox'].apply(lambda x: string_to_list(x)[0] <= bbox[0])) &
                (paths['bbox'].apply(lambda x: string_to_list(x)[1] >= bbox[1])) &
                (paths['bbox'].apply(lambda x: string_to_list(x)[2] >= bbox[2])) &
                (paths['bbox'].apply(lambda x: string_to_list(x)[3] <= bbox[3]))]
    elif data_source == 'gfs':
        path_list_predicate = paths[paths['path'].isin(choose_gfs(paths, time_start, time_end))]
        paths = paths[path_list_predicate['bbox'].apply(lambda x: string_to_list(x)[0] <= bbox[0]) &
                      (paths['bbox'].apply(lambda x: string_to_list(x)[1] >= bbox[1])) &
                      (paths['bbox'].apply(lambda x: string_to_list(x)[2] >= bbox[2])) &
                      (paths['bbox'].apply(lambda x: string_to_list(x)[3] <= bbox[3]))]
    candidate_tile_list = paths[paths['path'].apply(lambda x: ('_tile' in x) & (basin_id in x))]
    if len(candidate_tile_list) == 0:
        tile_list = generate_metadata(paths, data_source, bbox, time_start, time_end, basin_id)[0]
    else:
        tile_list = candidate_tile_list['path'][candidate_tile_list['bbox'].apply(lambda x: str(bbox) == x)].to_list()
        should_length = standard_length(data_source, time_start, time_end)
        if len(tile_list) < should_length - 2:
            tile_list = generate_metadata(paths, data_source, bbox, time_start, time_end, basin_id)[0]
    return tile_list


def generate_metadata(paths: DataFrame, data_source: str, bbox: list, time_start, time_end, basin_id):
    metadata_df = pd.read_csv(f's3://grids-origin/{data_source}_metadata.csv', storage_options=hdscc.MINIO_PARAM)
    tile_list = generate_tile_list(paths, data_source, bbox, time_start, time_end, basin_id)
    bbox_array = np.repeat(str(bbox), len(tile_list))
    time_start_array = np.repeat(str(time_start), len(tile_list))
    time_end_array = np.repeat(str(time_end), len(tile_list))
    if data_source != 'smap':
        res_array = np.repeat(hpm.get_para(data_source)[0], len(tile_list))
        temp_df = pd.DataFrame(
            {'bbox': bbox_array, 'time_start': time_start_array, 'time_end': time_end_array, 'res_lon': res_array,
             'res_lat': res_array, 'path': tile_list}, index=default_index(len(tile_list)))
    else:
        temp_df = pd.DataFrame(
            {'bbox': bbox_array, 'time_start': time_start_array, 'time_end': time_end_array,
             'res_lon': np.repeat(0.08, len(tile_list)), 'res_lat': np.repeat(0.08, len(tile_list)),
             'path': tile_list}, index=default_index(len(tile_list)))
    metadata_df = pd.concat([metadata_df, temp_df], axis=0)
    metadata_df.to_csv(f'{data_source}_metadata.csv', index=False)
    hdscc.FS.put_file(f'{data_source}_metadata.csv', f's3://grids-origin/{data_source}_metadata.csv')
    os.remove(f'{data_source}_metadata.csv')
    return tile_list, metadata_df


def generate_tile_list(paths: DataFrame, data_source, bbox, time_start, time_end, basin_id):
    tile_list = []
    for path in paths['path']:
        tile_path = path.rstrip('.nc4') + f'{basin_id}_tile.nc4'
        if data_source == 'smap':
            path_ds = h5py.File(hdscc.FS.open(path))
            # datetime.fromisoformat('2000-01-01T12:00:00') + timedelta(seconds=path_ds['time'][0])
            lon_array = path_ds['cell_lon'][0]
            lat_array = path_ds['cell_lat'][:, 0]
            cell_lon_w = np.argwhere(lon_array >= bbox[0])[0][0]
            cell_lon_e = np.argwhere(lon_array <= bbox[1])[-1][0]
            cell_lat_n = np.argwhere(lat_array <= bbox[2])[0][0]
            cell_lat_s = np.argwhere(lat_array >= bbox[3])[-1][0]
            tile_da = path_ds['Geophysical_Data']['sm_surface'][cell_lat_n: cell_lat_s + 1,
                      cell_lon_w: cell_lon_e + 1]
            tile_ds = xr.DataArray(tile_da).to_dataset(name='sm_surface')
        elif (data_source == 'era5_land') | (data_source == 'era5'):
            path_ds = access_fs.spec_path(path.lstrip('s3://'), head='minio')
            tile_ds = path_ds.sel(time=slice(time_start, time_end), longitude=slice(bbox[0], bbox[1]),
                                  latitude=slice(bbox[2], bbox[3]))
        else:
            # 会扰乱桶，注意
            path_ds = xr.open_dataset(hdscc.FS.open(path))
            if data_source == 'gpm':
                tile_ds = path_ds.sel(time=slice(time_start, time_end), lon=slice(bbox[0], bbox[1]),
                                      lat=slice(bbox[3], bbox[2]))
            # 会扰乱桶，注意
            elif data_source == 'gfs':
                tile_ds = path_ds.sel(time=slice(time_start, time_end), longitude=slice(bbox[0], bbox[1]),
                                      latitude=slice(bbox[2], bbox[3]))
            else:
                tile_ds = path_ds
        if data_source in ['gpm', 'smap', 'era5_land', 'era5']:
            hdscc.FS.write_bytes(tile_path, tile_ds.to_netcdf())
        elif data_source == 'gfs':
            tile_ds.to_netcdf('temp.nc4')
            hdscc.FS.put_file('temp.nc4', tile_path)
            os.remove('temp.nc4')
        tile_list.append(tile_path)
    return tile_list


def standard_length(data_source, time_start, time_end):
    if data_source == 'gpm':
        time_length = len(pd.date_range(time_start, time_end, freq='30min').to_list())
    elif data_source == 'smap':
        time_length = len(pd.date_range(time_start, time_end, freq='3h').to_list())
    elif (data_source == 'gfs') | (data_source == 'era5_land') | (data_source == 'era5'):
        time_length = len(pd.date_range(time_start, time_end, freq='h').to_list())
    else:
        time_length = len(pd.date_range(time_start, time_end, freq='d').to_list())
    return time_length


def choose_gfs(paths, start_time, end_time):
    """
        This function chooses GFS data within a specified time range and bounding box.
        Args:
            paths (Dataframe): A list of GFS data paths.
            start_time (datetime, YY-mm-dd): The start time of the desired data.
            end_time (datetime, YY-mm-dd): The end time of the desired data.
        Returns:
            list: A list of GFS data within the specified time range and bounding box.
        """
    path_list = []
    produce_times = ['00', '06', '12', '18']
    if start_time is None:
        start_time = paths['time_start'].iloc[0]
    if end_time is None:
        end_time = paths['time_end'].iloc[-1]
    time_range = pd.date_range(start_time, end_time, freq='1D')
    for date in time_range:
        date_str = date.strftime('%Y/%m/%d')
        for i in range(len(produce_times)):
            for j in range(6 * i, 6 * (i + 1)):
                path = 's3://grids-origin/GFS/GEE/1h/' + date_str + '/' + produce_times[i] + '/gfs20220103.t' + \
                       produce_times[i] + 'z.nc4.0p25.f' + '{:03d}'.format(j)
                path_list.append(path)
    return path_list


def string_to_list(x: str):
    return list(map(float, x[1:-1].split(',')))


def generate_bbox_from_shp(basin_shape_path, data_source, minio=True):
    # 只考虑单个流域
    if minio:
        basin_gpd = access_fs.spec_path(basin_shape_path.lstrip('s3://'), head="minio")
    else:
        basin_gpd = gpd.read_file(basin_shape_path)
    if data_source == 'smap':
        # 西，东，北，南
        bounds_array = basin_gpd.bounds.to_numpy()[0]
        bbox = [bounds_array[0], bounds_array[2], bounds_array[3], bounds_array[1]]
    else:
        mask = hpm.gen_single_mask(basin_shape_path.split('/')[-1], basin_gpd, dataname=data_source)
        bbox = [mask['lon'].values.min(), mask['lon'].values.max(), mask['lat'].values.max(), mask['lat'].values.min()]
    return bbox, basin_gpd


def grid_mean_mask(basin_id, times: list, data_source):
    # basin_id: basin_CHN_songliao_21401550, 碧流河
    # times: [[2023-06-06 00:00:00, 2023-06-06 02:00:00], [2023-06-07 00:00:00, 2023-06-07 02:00:00]]
    basin_shp = f's3://basins-origin/basin_shapefiles/{basin_id}.zip'
    bbox, basin = generate_bbox_from_shp(basin_shp, data_source=data_source)
    aver_npy = f'{basin_id}_{times}_{data_source}_hour_array.npy'
    s3_aver_npy_path = f's3://basins-origin/hour_data/1h/mean_data/{aver_npy}'
    if hdscc.FS.exists(s3_aver_npy_path):
        result_arr_list = np.load(hdscc.FS.open(s3_aver_npy_path), allow_pickle=True)
    else:
        aoi_data_paths = []
        for time_slice in times:
            time_start = time_slice[0]
            time_end = time_slice[1]
            aoi_path = query_path_from_metadata(basin_id, time_start, time_end, bbox, data_source=data_source)
            aoi_data_paths.append(aoi_path)
        result_arr_list = []
        for time_paths in aoi_data_paths:
            for path in time_paths:
                aoi_dataset = xr.open_dataset(hdscc.FS.open(path))
                mask = hpm.gen_single_mask(basin_id, basin, data_source)
                if data_source == 'gpm':
                    # 按照mask出来的四至和get_para()有关，全是.0或.5，直接输入四至就会破坏这样的性质
                    result_arr = hpm.mean_by_mask(aoi_dataset, var='precipitationCal', mask=mask)
                elif (data_source == 'era5_land') | (data_source == 'era5'):
                    result_arr = hpm.mean_by_mask(aoi_dataset, var='tp', mask=mask)
                elif data_source == 'smap':
                    result_arr = hpm.mean_by_mask(aoi_dataset, var='sm_surface', mask=mask)
                elif data_source == 'gfs':
                    result_arr = hpm.mean_by_mask(aoi_dataset, var='APCP', mask=mask)
                else:
                    result_arr = []
                result_arr_list.append(result_arr)
        result_arr_list = np.array(result_arr_list)
        np.save(aver_npy, result_arr_list)
        hdscc.FS.put_file(aver_npy, s3_aver_npy_path)
        os.remove(aver_npy)
    return result_arr_list, basin


def concat_gpm_average(basin_id, times: list):
    gpm_aver_npy = f'{basin_id}_{times}_gpm_hour_array_concat.npy'
    s3_gpm_aver_path = 's3://basins-origin/hour_data/1h/mean_data/' + gpm_aver_npy
    if hdscc.FS.exists(s3_gpm_aver_path):
        gpm_hour_array = np.load(hdscc.FS.open(s3_gpm_aver_path), allow_pickle=True)
    else:
        gpm_half_hour_array, basin = grid_mean_mask(basin_id, times, 'gpm')
        gpm_hour_array = []
        for i in np.arange(0, len(gpm_half_hour_array) - 1, 2):
            gpm_hour_i = np.add(gpm_half_hour_array[i], gpm_half_hour_array[i + 1])
            gpm_hour_array.append(gpm_hour_i)
        # temporarily fix
        if len(gpm_hour_array) % 2 != 0:
            gpm_hour_array.extend(gpm_hour_array[-1])
        gpm_hour_array = np.array(gpm_hour_array, dtype=object)
        np.save(gpm_aver_npy, gpm_hour_array)
        hdscc.FS.put_file(gpm_aver_npy, s3_gpm_aver_path)
        os.remove(gpm_aver_npy)
    return gpm_hour_array


def concat_gpm_smap_mean_data(basin_ids: list, times: list, use_pp=False):
    pp_sta_gdf = gpd.read_file(hdscc.FS.open('s3://stations-origin/stations_list/pp_stations.zip'))
    merge_list = []
    for basin_id in basin_ids:
        gpm_mean = concat_gpm_average(basin_id, times)
        smap_mean_mask, basin_gdf = grid_mean_mask(basin_id, times, 'smap')
        smap_mean = np.repeat(smap_mean_mask, 3, axis=1).flatten()
        if smap_mean.shape[0] < gpm_mean.shape[0]:
            diff = gpm_mean.shape[0] - smap_mean.shape[0]
            smap_mean = np.pad(smap_mean, (0, diff), 'edge')
        else:
            smap_mean = smap_mean[:gpm_mean.shape[0]]
        streamflow_arr = (read_streamflow_from_minio(times, basin_id.lstrip('basin_')))['Q'].to_numpy()
        if streamflow_arr.shape[0] < gpm_mean.shape[0]:
            diff = gpm_mean.shape[0] - streamflow_arr.shape[0]
            streamflow_arr = np.pad(streamflow_arr, (0, diff), 'edge')
        if use_pp:
            pp_stas_basin = gpd.overlay(pp_sta_gdf, basin_gdf, 'intersection')
            average_rainfall = rainfall_average(basin_gdf, pp_stas_basin, pp_stas_basin['ID'].to_list(), times[0][0])
            average_rainfall_times = average_rainfall[average_rainfall['TM'].isin(convert_time_slice_to_range(times))]
            temp_df = pd.DataFrame({'time': convert_time_slice_to_range(times), 'gpm_tp(mm/h)': gpm_mean, 'smap(m3/m3)': smap_mean,
                                    'sta_tp(mm/h)': average_rainfall_times['weighted_rainfall'].to_numpy(),
                                    'streamflow(m3/s)': streamflow_arr}).set_index(keys=['time'])
        else:
            temp_df = pd.DataFrame({'time': convert_time_slice_to_range(times), 'gpm_tp(mm/h)': gpm_mean, 'smap(m3/m3)': smap_mean,
                                    'streamflow(m3/s)': streamflow_arr}).set_index(keys=['time'])
        merge_ds = xr.Dataset.from_dataframe(temp_df.astype('float64'))
        merge_list.append(merge_ds)
        hdscc.FS.write_bytes(f's3://basins-origin/hour_data/1h/mean_data/mean_data_merged/mean_data_{basin_id}.nc',
                             merge_ds.to_netcdf())
    return merge_list


def convert_time_slice_to_range(time_slice_list):
    time_range_list = []
    for time_slice in time_slice_list:
        pd_time_slice = pd.date_range(time_slice[0], time_slice[1], freq='H').to_list()
        time_range_list.extend(pd_time_slice)
    return time_range_list


def read_streamflow_from_minio(times: list, sta_id=''):
    # sta_id: CHN_songliao_21401550、USA_xxx_01301500
    stream_times_df = pd.DataFrame()
    for time_slice in times:
        if (pd.to_datetime('2020-01-01') > pd.to_datetime(time_slice[1])) & ('camels' in sta_id):
            stcd = sta_id.split('_')[-1]
            zq_1h_path = f'datasets-origin/camels-hourly/data/usgs_streamflow_csv/{stcd}-usgs-hourly.csv'
            if hdscc.FS.exists(zq_1h_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zq_1h_path), index_col=None, parse_dates=['date'])
                streamflow_df = streamflow_df[streamflow_df['date'].isin(pd.date_range(time_slice[0], time_slice[1], freq='H'))]
                streamflow_df = streamflow_df.rename(columns={'date': 'TM', 'QObs(mm/h)': 'Q'})
                streamflow_df = streamflow_df[['TM', 'Q']]
            else:
                streamflow_df = pd.DataFrame()
        else:
            if 'camels' in sta_id:
                stcd = sta_id.split('_')[-1]
                sta_id = f'USA_usgs_{stcd}'
            zq_1h_path = f's3://stations-origin/zq_stations/hour_data/1h/zq_{sta_id}.csv'
            zq_6h_path = f's3://stations-origin/zq_stations/hour_data/6h/zq_{sta_id}.csv'
            zq_1d_path = f's3://stations-origin/zq_stations/day_data/1d/zq_{sta_id}.csv'
            zz_1h_path = f's3://stations-origin/zz_stations/hour_data/1h/zz_{sta_id}.csv'
            zz_6h_path = f's3://stations-origin/zz_stations/hour_data/6h/zz_{sta_id}.csv'
            zz_1d_path = f's3://stations-origin/zz_stations/day_data/1d/zz_{sta_id}.csv'
            if hdscc.FS.exists(zq_1h_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zq_1h_path), index_col=None, parse_dates=['TM'])
            elif hdscc.FS.exists(zq_6h_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zq_6h_path), index_col=None, parse_dates=['TM'])
            elif hdscc.FS.exists(zz_1h_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zz_1h_path), index_col=None, parse_dates=['TM'])
            elif hdscc.FS.exists(zz_6h_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zz_6h_path), index_col=None, parse_dates=['TM'])
            elif hdscc.FS.exists(zz_1d_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zz_1d_path), index_col=None, parse_dates=['TM'])
            elif hdscc.FS.exists(zq_1d_path):
                streamflow_df = pd.read_csv(hdscc.FS.open(zq_1d_path), index_col=None, parse_dates=['TM'])
            else:
                streamflow_df = pd.DataFrame()
            streamflow_df = streamflow_df[streamflow_df['TM'].isin(pd.date_range(time_slice[0], time_slice[1], freq='H'))]
            streamflow_df = streamflow_df[['TM', 'Q']]
        stream_times_df = pd.concat([stream_times_df, streamflow_df])
    return stream_times_df


'''
def merge_with_spatial_average(gpm_file, gfs_file, smap_file, output_file_path):
    def calculate_and_rename(input_file_path, prefix):
        ds = access_fs.spec_path(input_file_path, head="minio")
        avg_ds = ds.mean(dim=["lat", "lon"], skipna=True).astype("float32")
        new_names = {var_name: (
            f"{prefix}_tp" if var_name in ["tp", "__xarray_dataarray_variable__"] else f"{prefix}_{var_name}") for
            var_name in avg_ds.data_vars}
        avg_ds_renamed = avg_ds.rename(new_names)
        return avg_ds_renamed
    basin_id = output_file_path.split('_')[-1].split('.')[0]
    gfs_avg_renamed = calculate_and_rename(gfs_file, "gfs")
    gpm_avg_renamed = calculate_and_rename(gpm_file, "gpm")
    smap_avg_renamed = calculate_and_rename(smap_file, "smap")
    intersect_time = np.intersect1d(gfs_avg_renamed.time.values, gpm_avg_renamed.time.values, assume_unique=True)
    intersect_time = np.intersect1d(intersect_time, smap_avg_renamed.time.values, assume_unique=True)
    gfs_intersected = gfs_avg_renamed.sel(time=intersect_time)
    gpm_intersected = gpm_avg_renamed.sel(time=intersect_time)
    smap_intersected = smap_avg_renamed.sel(time=intersect_time)
    merged_ds = xr.merge([gfs_intersected, gpm_intersected, smap_intersected])
    merged_ds = merged_ds.assign_coords({"basin": basin_id}).expand_dims("basin")
    conf.FS.write_bytes(output_file_path, merged_ds.to_netcdf())
    return merged_ds
'''



import re
import numpy as np
from netCDF4 import Dataset, date2num, num2date
import time
from datetime import datetime, timedelta
import pandas as pd
import pint
import xarray as xr

# please don't remove the following line although it seems not used
import pint_xarray  # noqa


def creatspinc(value, data_vars, lats, lons, starttime, filename, resolution):
    gridspi = Dataset(filename, "w", format="NETCDF4")

    # dimensions
    gridspi.createDimension("time", value[0].shape[0])
    gridspi.createDimension("lat", value[0].shape[2])  # len(lat)
    gridspi.createDimension("lon", value[0].shape[1])

    # Create coordinate variables for dimensions
    times = gridspi.createVariable("time", np.float64, ("time",))
    latitudes = gridspi.createVariable("lat", np.float32, ("lat",))
    longitudes = gridspi.createVariable("lon", np.float32, ("lon",))

    # Create the actual variable
    for var, attr in data_vars.items():
        gridspi.createVariable(
            var,
            np.float32,
            (
                "time",
                "lon",
                "lat",
            ),
        )

    # Global Attributes
    gridspi.description = "var"
    gridspi.history = f"Created {time.ctime(time.time())}"
    gridspi.source = "netCDF4 python module tutorial"

    # Variable Attributes
    latitudes.units = "degree_north"
    longitudes.units = "degree_east"
    times.units = "days since 1970-01-01 00:00:00"
    times.calendar = "gregorian"

    # data
    latitudes[:] = lats
    longitudes[:] = lons

    # Fill in times
    dates = []
    if resolution == "daily":
        for n in range(value[0].shape[0]):
            dates.append(starttime + n)
        times[:] = dates[:]

    elif resolution == "6-hourly":
        # for n in range(value[0].shape[0]):
        #     dates.append(starttime + (n+1) * np.timedelta64(6, 'h'))

        for n in range(value[0].shape[0]):
            dates.append(starttime + (n + 1) * timedelta(hours=6))

        times[:] = date2num(dates, units=times.units, calendar=times.calendar)
        # print 'time values (in units %s): ' % times.units +'\n', times[:]
        dates = num2date(times[:], units=times.units, calendar=times.calendar)

    # Fill in values
    i = 0
    for var, attr in data_vars.items():
        gridspi.variables[var].long_name = attr["long_name"]
        gridspi.variables[var].units = attr["units"]
        gridspi.variables[var][:] = value[i][:]
        i = i + 1

    gridspi.close()


def regen_box(bbox, resolution, offset):
    lx = bbox[0]
    rx = bbox[2]
    LLON = np.round(
        int(lx)
        + resolution * int((lx - int(lx)) / resolution + 0.5)
        + offset
        * (int(lx * 10) / 10 + offset - lx)
        / abs(int(lx * 10) // 10 + offset - lx + 0.0000001),
        3,
    )
    RLON = np.round(
        int(rx)
        + resolution * int((rx - int(rx)) / resolution + 0.5)
        - offset
        * (int(rx * 10) / 10 + offset - rx)
        / abs(int(rx * 10) // 10 + offset - rx + 0.0000001),
        3,
    )

    by = bbox[1]
    ty = bbox[3]
    BLAT = np.round(
        int(by)
        + resolution * int((by - int(by)) / resolution + 0.5)
        + offset
        * (int(by * 10) / 10 + offset - by)
        / abs(int(by * 10) // 10 + offset - by + 0.0000001),
        3,
    )
    TLAT = np.round(
        int(ty)
        + resolution * int((ty - int(ty)) / resolution + 0.5)
        - offset
        * (int(ty * 10) / 10 + offset - ty)
        / abs(int(ty * 10) // 10 + offset - ty + 0.0000001),
        3,
    )

    # print(LLON,BLAT,RLON,TLAT)
    return [LLON, BLAT, RLON, TLAT]


def validate(date_text, formatter, error):
    try:
        return datetime.strptime(date_text, formatter)
    except ValueError as e:
        raise ValueError(error) from e


def cf2datetime(ds):
    ds = ds.copy()
    time_tmp1 = ds.indexes["time"]
    attrs = ds.coords["time"].attrs
    time_tmp2 = []
    for i in range(time_tmp1.shape[0]):
        tmp = time_tmp1[i]
        a = str(tmp.year).zfill(4)
        b = str(tmp.month).zfill(2)
        c = str(tmp.day).zfill(2)
        d = str(tmp.hour).zfill(2)
        e = str(tmp.minute).zfill(2)
        f = str(tmp.second).zfill(2)
        time_tmp2.append(np.datetime64(f"{a}-{b}-{c} {d}:{e}:{f}.00000000"))
    ds = ds.assign_coords(time=time_tmp2)
    ds.coords["time"].attrs = attrs

    return ds


def generate_time_intervals(start_date, end_date):
    # Initialize an empty list to store the intervals
    intervals = []

    # Loop over days
    while start_date <= end_date:
        # Loop over the four time intervals in a day
        intervals.extend(
            [start_date.strftime("%Y-%m-%d"), hour] for hour in ["00", "06", "12", "18"]
        )
        # Move to the next day
        start_date += timedelta(days=1)

    return intervals


def _convert_target_unit(target_unit):
    """Convert user-friendly unit to standard unit for internal calculations."""
    if match := re.match(r"mm/(\d+)(h|d)", target_unit):
        num, unit = match.groups()
        return int(num), unit
    return None, None


def streamflow_unit_conv(streamflow, area, target_unit="mm/d", inverse=False):
    """Convert the unit of streamflow data to mm/xx(time) for a basin or inverse.

    Parameters
    ----------
    streamflow: xarray.Dataset, numpy.ndarray, pandas.DataFrame/Series
        Streamflow data of each basin.
    area: xarray.Dataset or pint.Quantity wrapping numpy.ndarray, pandas.DataFrame/Series
        Area of each basin.
    target_unit: str
        The unit to convert to.
    inverse: bool
        If True, convert the unit to m^3/s.
        If False, convert the unit to mm/day or mm/h.

    Returns
    -------
    Converted data in the same type as the input streamflow.
    """
    # Convert the user input unit format
    num, unit = _convert_target_unit(target_unit)
    if unit:
        if unit == "h":
            standard_unit = "mm/h"
            conversion_factor = num
        elif unit == "d":
            standard_unit = "mm/d"
            conversion_factor = num
        else:
            raise ValueError(f"Unsupported unit: {unit}")
    else:
        standard_unit = target_unit
        conversion_factor = 1
    # Regular expression to match units with numbers
    custom_unit_pattern = re.compile(r"mm/(\d+)(h|d)")

    # Function to handle the conversion for numpy and pandas
    def np_pd_conversion(streamflow, area, target_unit, inverse, conversion_factor):
        if not inverse:
            result = (streamflow / area).to(target_unit) * conversion_factor
        else:
            result = (streamflow * area).to(target_unit) / conversion_factor
        return result.magnitude

    # Handle xarray
    if isinstance(streamflow, xr.Dataset) and isinstance(area, xr.Dataset):
        streamflow_units = streamflow[list(streamflow.keys())[0]].attrs.get(
            "units", None
        )
        if not inverse:
            if not (
                custom_unit_pattern.match(target_unit)
                or re.match(r"mm/(?!\d)", target_unit)
            ):
                raise ValueError(
                    "target_unit should be a valid unit like 'mm/d', 'mm/day', 'mm/h', 'mm/hour', 'mm/3h', 'mm/5d'"
                )

            q = streamflow.pint.quantify()
            a = area.pint.quantify()
            r = q[list(q.keys())[0]] / a[list(a.keys())[0]]
            # result = r.pint.to(target_unit).to_dataset(name=list(q.keys())[0])
            result = (r.pint.to(standard_unit) * conversion_factor).to_dataset(
                name=list(q.keys())[0]
            )
            # Manually set the unit attribute to the custom unit
            result_ = result.pint.dequantify()
            result_[list(result_.keys())[0]].attrs["units"] = target_unit
            return result_
        else:
            if streamflow_units:
                if custom_match := custom_unit_pattern.match(streamflow_units):
                    num, unit = custom_match.groups()
                    if unit == "h":
                        standard_unit = "mm/h"
                        conversion_factor = int(num)
                    elif unit == "d":
                        standard_unit = "mm/d"
                        conversion_factor = int(num)
                    # Convert custom unit to standard unit
                    r_ = streamflow / conversion_factor
                    r_[list(r_.keys())[0]].attrs["units"] = standard_unit
                    r = r_.pint.quantify()
                else:
                    r = streamflow.pint.quantify()
            else:
                r = streamflow.pint.quantify()
            if target_unit not in ["m^3/s", "m3/s"]:
                raise ValueError("target_unit should be 'm^3/s'")
            a = area.pint.quantify()
            q = r[list(r.keys())[0]] * a[list(a.keys())[0]]
            result = q.pint.to(target_unit).to_dataset(name=list(r.keys())[0])
            # dequantify to get normal xr_dataset
            return result.pint.dequantify()

    # Handle numpy and pandas
    elif isinstance(streamflow, pint.Quantity) and isinstance(area, pint.Quantity):
        if type(streamflow.magnitude) not in [np.ndarray, pd.DataFrame, pd.Series]:
            raise TypeError(
                "Input streamflow must be xarray.Dataset, or pint.Quantity wrapping numpy.ndarray, or pandas.DataFrame/Series"
            )
        if type(area.magnitude) != type(streamflow.magnitude):
            raise TypeError("streamflow and area must be the same type")
        return np_pd_conversion(
            streamflow, area, standard_unit, inverse, conversion_factor
        )

    else:
        raise TypeError(
            "Input streamflow must be xarray.Dataset, or pint.Quantity wrapping numpy.ndarray, or pandas.DataFrame/Series"
        )



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-22 18:02:00
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-04-28 10:37:20
FilePath: /hydrodatasource/tests/test_rainfall_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

import pytest
from hydrodatasource.cleaner.rainfall_cleaner import RainfallCleaner
import pandas as pd
import matplotlib.pyplot as plt


def test_anomaly_process():
    # 测试降雨数据处理功能
    cleaner = RainfallCleaner(
         data_path= "/home/liutianxv1/降雨sampledatatest.csv",  era5_path="/home/liutianxv1/era5数据读取/",
        station_file="/home/liutianxv1/pp_stations.csv",start_time='2020-01-01',end_time='2022-10-07'
    )
    # methods默认可以联合调用，也可以单独调用。大多数情况下，默认调用detect_sum
    methods = ["detect_era5"]
    cleaner.anomaly_process(methods)

    print(cleaner.origin_df)
    print(cleaner.processed_df)
    # cleaner.processed_df.to_csv("/home/liutianxv1/降雨sampledatatest.csv")
    # cleaner.temporal_list.to_csv("/home/liutianxv1/降雨temporal_list.csv")
    cleaner.spatial_list.to_csv("/home/liutianxv1/降雨spatial_list.csv")



"""
Author: Jeff Zhu
Date: 2023-10-06 20:50:41
LastEditTime: 2023-10-14 19:30:14
LastEditors: Wenyu Ouyang
Description: Top-level package for wis-downloader
FilePath: \hydrodata\hydrodata\downloader\__init__.py
Copyright (c) 2023-2024 Jeff Zhu. All rights reserved.
"""

__author__ = """Jeff Zhu"""
__email__ = "zjf014@gmail.com"
__version__ = "0.0.1"



import glob
import logging
import os

import geopandas as gpd
import intake as intk
import pandas as pd
import ujson
import xarray as xr
import zarr
from kerchunk.hdf import SingleHdf5ToZarr
import hydrodatasource.configs.config as conf


def spec_path(url_path: str, head='local', need_cache=False, is_dir=False):
    if is_dir is False:
        if head == 'local':
            url_path = os.path.join(conf.LOCAL_DATA_PATH, url_path)
            ret_data = read_valid_data(url_path, need_cache=need_cache)
        elif head == 'minio':
            url_path = 's3://' + url_path
            ret_data = read_valid_data(url_path, storage_option=conf.MINIO_PARAM, need_cache=need_cache)
        else:
            raise ValueError("head should be 'local' or 'minio'")
    else:
        ret_data = []
        if head == 'local':
            url_path = os.path.join(conf.LOCAL_DATA_PATH, url_path)
            for file in glob.glob(url_path + '/**', recursive=True):
                if not os.path.isdir(file):
                    ret_data.append(read_valid_data(file, need_cache=need_cache))
        elif head == 'minio':
            url_path = 's3://' + url_path
            for file in conf.FS.glob(url_path + '/**'):
                if not conf.FS.isdir(file):
                    ret_data.append(read_valid_data(file, storage_option=conf.MINIO_PARAM, need_cache=need_cache))
        else:
            raise ValueError("head should be 'local' or 'minio'")
    return ret_data


# Extract from HydroForecast
def read_valid_data(obj: str, storage_option=None, need_cache=False, need_refer=False):
    """
    Read valid data from different file types.
    See https://intake.readthedocs.io/en/latest/plugin-directory.html
    pip install intake-xarray intake-geopandas

    Parameters:
    obj (str): The file path or URL of the data, format is 's3://bucket_name/file_name'.
    storage_option (dict, optional): The storage options for accessing the data. Defaults to None.
    need_cache (bool, optional): Whether to cache the data. Defaults to False.

    Returns:
    object: The data object.
    """
    data_obj = None
    dot_in_obj = '.' in obj
    cache_name = obj.lstrip('s3://').split('/')[-1]
    if not dot_in_obj:
        data_obj = pd.read_fwf(obj, storage_options=storage_option)
        if (need_cache is True) & (storage_option is not None):
            data_obj.to_file(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
    elif dot_in_obj:
        ext_name = obj.split('.')[-1]
        if ext_name == 'csv':
            data_obj = pd.read_csv(obj, storage_options=storage_option)
            data_obj = xr.Dataset(data_obj)
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_csv(path_or_buf=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif (ext_name == 'nc') or (ext_name == 'nc4') or (ext_name == 'hdf5') or (ext_name == 'h5') or ('nc4' in obj):
            if need_refer is True:
                data_obj = gen_refer_and_read_zarr(obj, storage_option=storage_option)
            else:
                '''
                nc_source = intk.datatypes.HDF5(obj, storage_options=storage_option)
                nc_src_reader = intk.readers.DaskHDF(nc_source).to_reader()
                data_obj: xr.Dataset = nc_src_reader.read()
                '''
                if (ext_name == 'nc') or (ext_name == 'nc4') or ('nc4' in obj):
                    data_obj = xr.open_dataset(conf.FS.open(obj), chunks='auto')
                elif (ext_name == 'hdf5') or (ext_name == 'h5'):
                    data_obj = xr.open_dataset(conf.FS.open(obj), engine='h5netcdf', chunks='auto', phony_dims='access')
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_netcdf(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == 'json':
            data_obj = pd.read_json(obj, storage_options=storage_option)
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_json(path_or_buf=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == 'zip':
            # Now zipfile is used to read shapefile
            # Can't read shapefile directly, see this: https://github.com/geopandas/geopandas/issues/3129
            # pip install pyogrio
            data_obj = gpd.read_file(conf.FS.open(obj), engine='pyogrio')
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_file(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif 'grb2' in obj:
            # ValueError: unrecognized engine cfgrib must be one of: ['netcdf4', 'h5netcdf', 'scipy', 'store', 'zarr']
            # https://blog.csdn.net/weixin_44052055/article/details/108658464?spm=1001.2014.3001.5501
            # 似乎只能用conda来装eccodes
            grib_ds = xr.open_dataset(conf.FS.open(obj))
            if (need_cache is True) & (storage_option is not None):
                grib_ds.to_netcdf(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == 'txt':
            data_obj = pd.read_fwf(obj, storage_options=storage_option)
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_csv(os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == 'zarr':
            zarr_mapper = conf.FS.get_mapper(obj)
            # KVStore is introduced in zarr specification V3
            # https://zarr-specs.readthedocs.io/en/latest/v3/stores.html
            zarr_store = zarr.storage.KVStore(zarr_mapper)
            data_obj = xr.open_zarr(zarr_store)
        else:
            logging.error(f'Unsupported file type: {ext_name}')
    else:
        data_obj = object()
        logging.error("这是数据存储，不是百度云盘！")
    return data_obj


def gen_refer_and_read_zarr(obj_path, storage_option=None):
    # https://github.com/fsspec/kerchunk/discussions/431
    obj_json_path = 's3://references/' + str(obj_path) + '.json'
    if not conf.FS.exists(obj_json_path):
        obj_path = 's3://' + obj_path
        with conf.FS.open(obj_path, 'rb') as fpj:
            nc_chunks = SingleHdf5ToZarr(fpj, obj_path)
            with conf.FS.open(obj_json_path, 'wb') as fp:
                fp.write(ujson.dumps(nc_chunks.translate()).encode())
    data_type_obj = intk.datatypes.HDF5(obj_json_path, storage_options=storage_option)
    data_reader_obj = intk.readers.XArrayDatasetReader(data_type_obj).to_reader()
    data_obj = data_reader_obj.read()
    return data_obj

